# coding: utf-8

from typing import List, Tuple, Optional, Callable, Union, Dict, Any
import numpy as np
from pathlib import Path
from ..tool import DISVE as VE, DISVE_config, Spot_boxes
from .camera import Camera
from .displayer import Displayer


class DISVE(Camera):
  """This block tracks the motion of regions of an image (patches), taking the
  first image as a reference.

  It relies on cross-correlation, and is thus well-suited for tracking speckled
  patches. The displacement output may then be used to compute strains, and so
  to perform video-extensometry.

  The images may be acquired by a camera, or be sent from another block.
  Several algorithms are available for tracking the patches, with different
  characteristics. All the computations are done by the :ref:`Disve tool`.
  """

  def __init__(self,
               camera: str,
               patches: List[Tuple[int, int, int, int]],
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               config: bool = True,
               display_images: bool = False,
               displayer_backend: Optional[str] = None,
               displayer_framerate: float = 5,
               verbose: bool = False,
               freq: float = 200,
               save_images: bool = False,
               img_name: str = "{self._n_loops:6d}_{t-self.t0:.6f}.tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               labels: Optional[List[str]] = None,
               method: str = 'Disflow',
               alpha: float = 3,
               delta: float = 1,
               gamma: float = 0,
               finest_scale: int = 1,
               iterations: int = 1,
               gradient_iterations: int = 10,
               patch_size: int = 8,
               patch_stride: int = 3,
               border: float = 0.2,
               safe: bool = True,
               follow: bool = True,
               **kwargs) -> None:
    """Sets a few attributes.

    Args:
      camera: The camera to use for acquiring the images. It should be one of
        the :ref:`Supported cameras`.
      patches: A list containing the different patches to track. Each patch
        should be given as follows : ``(y min, x_min, height, width)``.
      transform: A function taking an image as an argument and returning a
        transformed image. The original image is discarded and only the
        transformed one is kept for processing, display and saving.
      config: If :obj:`True`, a config window is shown before the test starts
        for interactively tuning the camera settings. It also allows selecting
        the spots to track.
      display_images: If :obj:`True`, a window displays the acquired images
        in low resolution during the test. This display is mainly intended for
        debugging and visual follow-up, but not for displaying high-quality
        images.
      displayer_backend: If ``display_images`` is :obj:`True`, the backend to
        use for the display window. Should be one of :
        ::

          'cv2', 'mpl'

        If not given, OpenCV will be used if available.
      displayer_framerate: If ``display_images`` is :obj:`True`, sets the
        maximum framerate for updating the display window. This setting allows
        limiting the resources used by the displayer. Note that the actual
        achieved framerate might differ, this is just the maximum limit.
      verbose: If :obj:`True`, the achieved framerate will be displayed in the
        console during the test.
      freq: If given, the block will try to loop at this frequency. If it is
        lower than the framerate of the camera, frames will be dropped. This
        argument can be used for limiting the achieved framerate when the
        camera doesn't support framerate control.
      save_images: If :obj:`True`, the acquired images are saved on the
        computer during the test. Note that saving images uses CPU, so the
        achieved performance might drop when this feature is in use.
      img_name: If ``save_images`` is :obj:`True`, the template for naming the
        recorded images. It is evaluated as an `f-string`, and must contain the
        file extension at the end. For building the `f-string`, the
        ``self._n_loops`` attribute holds the loop number, and ``t-self.t0``
        holds the current timestamp.
      save_folder: If ``save_images`` is :obj:`True`, the directory to save
        images to. If it doesn't exist, it will be created. If not given, the
        images are saved in a folder named `Crappy_images` and created next to
        the file being run.
      save_period: If ``save_images`` is :obj:`True`, only one every this
        number of images will be saved.
      save_backend: The backend to use for saving the images. Should be one
        of :
        ::

          'sitk', 'pil', 'cv2'

        If not specified, SimpleITK will be used if available, then OpenCV as a
        second choice, and finally Pillow if none of the others was available.
      image_generator: A function taking two floats as arguments, and returning
        an image. It is only used for demonstration without camera in the
        examples, and isn't meant to be used in an actual test.
      labels: The labels associated with the timestamp and the displacement of
        the patches. If not given, the time label is ``t(s)``, the metadata
        label is ``'meta'``, the first patch displacement labels are ``p0x``
        and ``p0y``, the second ``p1x`` and ``p1y``, etc.
      method: The method to use to calculate the displacement. `Disflow` uses
        opencv's DISOpticalFlow and `Lucas Kanade` uses opencv's
        calcOpticalFlowPyrLK, while all other methods are based on a basic
        cross-correlation in the Fourier domain. `Pixel precision` calculates
        the displacement by getting the position of the maximum of the
        cross-correlation, and has thus a 1-pixel resolution. It is mainly
        meant for debugging. `Parabola` refines the result of
        `Pixel precision` by interpolating the neighborhood of the maximum, and
        have thus sub-pixel resolutions.
      alpha: Weight of the smoothness term in DisFlow.
      delta: Weight of the color constancy term in DisFlow.
      gamma: Weight of the gradient constancy term in DisFlow.
      finest_scale: Finest level of the Gaussian pyramid on which the flow
        is computed in DisFlow (`0` means full scale).
      iterations: Maximum number of gradient descent iterations in the
        patch inverse search stage in DisFlow.
      gradient_iterations: Maximum number of gradient descent iterations
        in the patch inverse search stage in DisFlow.
      patch_size: Size of an image patch for matching in DisFlow
        (in pixels).
      patch_stride: Stride between neighbor patches in DisFlow. Must be
        less than patch size.
      border: Crop the patch on each side according to this value before
        calculating the displacements. 0 means no cropping, 1 means the entire
        patch is cropped.
      safe: If :obj:`True`, checks whether the patches aren't exiting the
        image, and raises an error if that's the case.
      follow: It :obj:`True`, the patches will move to follow the displacement
        of the image.
      **kwargs: Any additional argument to pass to the camera.
    """

    self._patches = Spot_boxes()
    self._patches.set_spots(patches)

    super().__init__(camera=camera,
                     transform=transform,
                     config=False,
                     display_images=False,
                     displayer_backend=displayer_backend,
                     displayer_framerate=displayer_framerate,
                     software_trig_label=None,
                     verbose=verbose,
                     freq=freq,
                     save_images=save_images,
                     img_name=img_name,
                     save_folder=save_folder,
                     save_period=save_period,
                     save_backend=save_backend,
                     image_generator=image_generator,
                     **kwargs)

    self._ve = VE(patches=self._patches,
                  method=method,
                  alpha=alpha,
                  delta=delta,
                  gamma=gamma,
                  finest_scale=finest_scale,
                  iterations=iterations,
                  gradient_iterations=gradient_iterations,
                  patch_size=patch_size,
                  patch_stride=patch_stride,
                  border=border,
                  safe=safe,
                  follow=follow)

    self._config_dis = config

    # Setting the labels
    if labels is None:
      self.labels = ['t(s)', 'meta'] + [elt
                                        for i, _ in enumerate(patches)
                                        for elt in [f'p{i}x', f'p{i}y']]
    else:
      self.labels = labels

    # Initializing the displayer
    if display_images:
      self._displayer_dis = Displayer(f"Displayer {camera} "
                                      f"{Camera.cam_count[self._camera_name]}",
                                      displayer_framerate,
                                      displayer_backend)
    else:
      self._displayer_dis = None

  def prepare(self) -> None:
    """Opens the camera for acquiring images and displays the corresponding
    settings window."""

    super().prepare()

    if self._config_dis:
      config = DISVE_config(self._camera, self._patches)
      config.main()

    if self._displayer_dis is not None:
      self._displayer_dis.prepare()

  def begin(self) -> None:
    """Takes a first image from the camera and uses it to initialize the Disve
    tool."""

    _, img = self._camera.get_image()
    self._ve.set_img0(img)

  def finish(self) -> None:
    """Closes the Disve tool and the camera acquiring the images."""

    if self._displayer_dis is not None:
      self._displayer_dis.finish()

    super().finish()

  def _additional_loop(self, meta: Dict[str, Any], img: np.ndarray) -> None:
    """Simply calculates the displacement and sends it to downstream blocks."""

    ret = self._ve.calculate_displacement(img)
    self.send([meta['t(s)'], meta] + ret)

    if self._displayer_dis is not None:
      for patch in self._patches:
        self._draw_box(img, patch)

      self._displayer_dis.update(img)



# coding: utf-8

"""
This program is often used as the starting point when performing tests on the
"Biotens" machine.

It creates a new folder for each experiment and performs tensile tests using
videoextensometry.
"""

import time

import crappy

save_path = "biotens_data/"
timestamp = time.ctime()[:-5].replace(" ", "_")
save_path += timestamp + "/"

# Creating F sensor
effort = crappy.blocks.IOBlock("Comedi", channels=[0], gain=[-48.8],
                               labels=['t(s)', 'F(N)'])

# grapher
graph_effort = crappy.blocks.Grapher(('t(s)', 'F(N)'))
crappy.link(effort, graph_effort)

# and recorder
rec_effort = crappy.blocks.Recorder(save_path + "effort.csv")
crappy.link(effort, rec_effort)

# Quick hack to reset the position of the actuator
b = crappy.actuator.Biotens()
b.open()
b.reset_position()
b.set_position(5, 50)

# Creating Machine block...
biotens = crappy.blocks.Machine([{'type': 'biotens',
                                  'port': '/dev/ttyUSB0',
                                  'pos_label': 'position1',
                                  'cmd': 'cmd'}])
# ..graph...
# graph_pos = crappy.blocks.Grapher(('t(s)', 'position1'))
# crappy.link(biotens, graph_pos)
# ...and recorder
rec_pos = crappy.blocks.Recorder(save_path + 'position.csv')
crappy.link(biotens, rec_pos)

# To pilot the biotens
generator = crappy.blocks.Generator([{'type': 'constant',
                                      'condition': 'F(N)>90',
                                      'value': 5}], freq=100)
crappy.link(effort, generator)
crappy.link(generator, biotens)

# VideoExtenso
extenso = crappy.blocks.Video_extenso(camera="Xiapi", white_spots=False)

# Recorder
rec_extenso = crappy.blocks.Recorder(save_path + 'extenso.csv',
                                     labels=['t(s)', 'Exx(%)', 'Eyy(%)'])
crappy.link(extenso, rec_extenso)
# And grapher
graph_extenso = crappy.blocks.Grapher(('t(s)', 'Exx(%)'), ('t(s)', 'Eyy(%)'))
crappy.link(extenso, graph_extenso)

# And here we go !
crappy.start()



# coding: utf-8

from .modifier import Modifier
from typing import Optional, Tuple, Dict, Any, Union, List


class Trig_on_value(Modifier):
  """Modifier passing the data to the downstream only if the value carried by a
  given label matches a given set of accepted values.

  Mostly useful to trig blocks.
  """

  def __init__(self,
               label: str,
               values: Union[Any, Tuple[Any, ...], List[Any]]) -> None:
    """Sets the args and initializes the parent class.

    Args:
      label: The name of the label to monitor.
      values: The values of ``label`` for which the data will be transmitted.
        Can be a single value, a :obj:`list` or a :obj:`tuple`.
    """

    super().__init__()

    self._label = label
    if isinstance(values, list) or isinstance(values, tuple):
      self._values = values
    else:
      self._values = (values,)

  def evaluate(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """Checks if the value of ``label`` is in the predefined set of accepted
    values, and if so transmits the data."""

    if data[self._label] in self._values:
      return data



# coding: utf-8

"""
Example showing how to use the DISVE block.

It performs DIC on the specified patches, and returns the motion along `x` and
`y` in pixels. This example demonstrates the processing to measure strain using
this method.

!! Important !!
  The w and h values may need to be adjusted to the actual size of your image.
  Otherwise an error may be raised.

  It was not possible to set them automatically here while still keeping the
  script short and easily understandable.

Required hardware:
  - Any camera
"""

import crappy

if __name__ == "__main__":
  w, h = 640, 480
  ps = 100  # patch size (x and y)
  m = 100  # Margin

  # Patches are defined as such: (y, x, height, width)
  # x and y being the coordinates to the upper-left corner
  patches = [
      (m, w // 2 - ps // 2, ps, ps),  # Top
      (h // 2 - ps // 2, w - ps - m, ps, ps),  # Right
      (h - ps - m, w // 2 - ps // 2, ps, ps),  # Bottom
      (h // 2 - ps // 2, m, ps, ps)]  # Left

  ve = crappy.blocks.DISVE('Webcam', patches, verbose=True,
                           display_images=True)
  graphy = crappy.blocks.Grapher(('t(s)', 'p0y'), ('t(s)', 'p2y'))
  graphx = crappy.blocks.Grapher(('t(s)', 'p1x'), ('t(s)', 'p3x'))

  crappy.link(ve, graphx)
  crappy.link(ve, graphy)

  def compute_strain(d):
    d['Exx(%)'] = (d['p3x'] - d['p1x']) / (w - ps) * 100
    d['Eyy(%)'] = (d['p0y'] - d['p2y']) / (h - ps) * 100
    return d

  graphstrain = crappy.blocks.Grapher(('t(s)', 'Exx(%)'), ('t(s)', 'Eyy(%)'))
  crappy.link(ve, graphstrain, modifier=compute_strain)
  crappy.start()



# coding: utf-8

from time import time
import numpy as np
from typing import List, Dict, Any, Optional, Tuple
from dataclasses import dataclass, field
from itertools import chain
from .inout import InOut
from .._global import OptionalModule

try:
  from labjack import ljm
except ImportError:
  ljm = OptionalModule("ljm",
                       "Please install Labjack LJM and the ljm Python module")


@dataclass
class _Channel:
  """This class is a simple structure holding all the attributes a Labjack
  channel can have for streaming."""

  name: str

  address: Optional[int] = None
  gain: float = 1
  offset: float = 0
  make_zero: bool = False
  range: float = 10
  write_at_open: List[Tuple[str, Any]] = field(default_factory=list)

  def update(self, dic_in: Dict[str, Any]) -> None:
    """Updates the channel keys based on the user input."""

    for key, val in dic_in.items():
      if hasattr(self, key):
        setattr(self, key, val)

      # Handling the case when the user enters a wrong key
      else:
        print(f"[Labjack T7] Unknown channel key : {key}, ignoring it")


class T7_streamer(InOut):
  """This InOut object allows controlling a Labjack T7 device in stream mode.

  It can only acquire data on the `AIN` channels. For single point mode, and
  acquisition on all channels, use the :ref:`Labjack T7` InOut.

  Compared with single point acquisition, the streaming mode can achieve much
  higher data rates and has a much greater regularity in the frequency of the
  acquisition. However, fewer options are available and not all types of
  channels can be read in the streamer mode.

  For each channel, the voltage range can be tuned, and a gain and offset can
  be defined. Also, the user can decide whether the channel should be zeroed
  before starting the test or not.

  Important:
    The ``streamer`` argument of the IOBlock controlling this InOut must be set
    to :obj:`True` to enable streaming in this class. Otherwise, only single
    point acquisition can be performed.
  """

  def __init__(self,
               channels: List[Dict[str, Any]],
               device: str = 'ANY',
               connection: str = 'ANY',
               identifier: str = 'ANY',
               scan_rate: int = 100000,
               scan_per_read: int = 10000,
               resolution: int = 1) -> None:
    """Sets the args and initializes the parent class.

    Args:
      channels: A :obj:`list` of the channels to interface with on the Labjack.
        Each object in this list should be a :obj:`dict` representing a single
        channel, and whose keys provide information on the channel to use.
        Refer to the note below for more information on the possible keys.
      device: The type of Labjack to open. Possible values include :
        ::

          'ANY', 'T7', 'T4', 'DIGIT'

        Only tested with `'T7'` in Crappy.
      connection: The type of connection used for interfacing with the Labjack.
        Possible values include :
        ::

          'ANY', 'TCP', 'USB', 'ETHERNET', 'WIFI'

      identifier: Any extra information allowing to further identify the
        Labjack to open, like a serial number, an IP address, or a device name.
      scan_rate: The acquisition frequency in Hz for all channels. Note that
        the sample rate (`scan_rate * num of chan`) cannot exceed `100000`. If
        it is too high it will be lowered to the highest possible value.
      scan_per_read: The number of points to read at each loop.
      resolution: The resolution of the acquisition as an integer for all
        channels. Refer to Labjack documentation for more details. The higher
        this value the better the resolution, but the lower the speed. The
        possible range is either `1` to `8` or to `12` depending on the model.
        The default is `1`.

    Note:
      - ``channels`` keys:

        - name: The name of the channel to interface with, as written on the
          Labjack's case. Ex: `'AIN0'`. In streamer mode, only the `AIN`
          channels, i.e. the analog inputs, are available.

        - gain: The measured value will be modified in Crappy as follows :
          ::

            returned_value = gain * measured_value + offset

        - offset: The measured value will be modified in Crappy as follows :
          ::

            returned_value = gain * measured_value + offset

        - make_zero: If :obj:`True`, data will be acquired on this channel
          before the test starts, and a compensation value will be deduced
          so that the offset of this channel is `0`. **It will only take effect
          if the ``make_zero_delay`` argument of the :ref:`IOBlock` controlling
          the Labjack is set** !

        - range: The range of the acquisition in Volts. A range of `x` means
          that values can be read  between `-x` and `x` Volts. The possible
          values are :
          ::

            0.01, 0.1, 1, 10

    """

    super().__init__()

    if len(channels) * scan_rate > 100000:
      scan_rate = 100000 / len(channels)
      print(f"[Labjack T7] Warning ! scan_rate is too high! Sample rate cannot"
            f" exceed 100kS/s, lowering samplerate to {scan_rate} samples/s")

    self._device = device
    self._connection = connection
    self._identifier = identifier
    self._scan_rate = scan_rate
    self._scan_per_read = scan_per_read
    self._resolution = resolution

    self._channels = list()

    # Parsing the setting dict given for each channel
    for channel in channels:

      # Checking that the name was given as it's the most important attribute
      if 'name' not in channel:
        raise AttributeError("[Labjack T7] The given channels must contain "
                             "the 'name' key !")

      # Instantiating the channel and its attributes
      chan = _Channel(name=channel['name'])
      chan.update(channel)
      chan.write_at_open.append((f"{chan.name}_RANGE", chan.range))
      chan.address, _ = ljm.nameToAddress(chan.name)

      self._channels.append(chan)

    # these attributes will be set later
    self._handle = None
    self._n_points = 0
    self._stream_t0 = 0
    self._stream_started = False

  def open(self) -> None:
    """Opens the Labjack, parses the commands to write at open, and sends them.

    Also checks whether the scan rate chose nby the Labjack is the same as
    requested by the user.
    """

    # Opening the Labjack
    self._handle = ljm.openS(self._device, self._connection, self._identifier)

    # Setting the different channels to read from on the Labjack
    write_at_open = list(chain(*(chan.write_at_open
                                 for chan in self._channels)))
    write_at_open.extend([("STREAM_SCANRATE_HZ", self._scan_rate),
                          ("STREAM_RESOLUTION_INDEX", self._resolution)])
    names, values = tuple(zip(*write_at_open))
    ljm.eWriteNames(handle=self._handle,
                    numFrames=len(names),
                    aNames=names,
                    aValues=values)

    # Checking if the scan rate that will be used is the same as requested
    scan_rate = ljm.eReadName(handle=self._handle, name="STREAM_SCANRATE_HZ")
    if scan_rate != self._scan_rate:
      print(f"[Labjack T7] Actual scan_rate: {scan_rate} "
            f"instead of {self._scan_rate}")
      self._scan_rate = scan_rate

  def make_zero(self, delay: float) -> None:
    """Overriding of the method of the parent class, because the user can
    choose which channels should be zeroed or not.

    It simply performs the regular zeroing, and resets the compensation for the
    channels that shouldn't be zeroed.
    """

    # No need to acquire data if no channel should be zeroed
    if any(chan.make_zero for chan in self._channels):

      # Acquiring the data
      super().make_zero(delay)

      # Proceed only if the acquisition went fine
      if self._compensations:

        # Resetting the compensation for channels that shouldn't be zeroed
        self._compensations = [comp if chan.make_zero else 0 for comp, chan
                               in zip(self._compensations, self._channels)]

  def start_stream(self) -> None:
    """Starts the stream, and saves the timestamp chen the stream started."""

    ljm.eStreamStart(handle=self._handle,
                     scansPerRead=self._scan_per_read,
                     numAddresses=len(self._channels),
                     aScanList=[chan.address for chan in self._channels],
                     scanRate=self._scan_rate)
    self._stream_t0 = time()
    self._stream_started = True

  def get_data(self) -> List[float]:
    """Reads single data points, applies the given gains and offsets, and
    returns the data."""

    data = ljm.eReadNames(handle=self._handle,
                          numFrames=len(self._channels),
                          aNames=[chan.name for chan in self._channels])

    return [time()] + [val * chan.gain + chan.offset for chan, val
                       in zip(self._channels, data)]

  def get_stream(self) -> Optional[List[np.ndarray]]:
    """Acquires the stream, reshapes the data, applies the gains and offsets,
    and returns the data along with a time array."""

    if not self._stream_started:
      return

    # Acquiring the data from the Labjack and reshaping it
    raw_data, *_ = ljm.eStreamRead(self._handle)
    data = np.array(raw_data)
    data = data.reshape(len(data) // len(self._channels), len(self._channels))

    # Applying the given gains and offsets
    for i, chan in enumerate(self._channels):
      data[:, i] = chan.gain * data[:, i] + chan.offset

    # Generating the array of time values
    t = self._stream_t0 + np.arange(self._n_points, self._n_points +
                                    data.shape[0]) / self._scan_rate
    self._n_points += data.shape[0]

    return [t[:, np.newaxis], data]

  def stop_stream(self) -> None:
    """Stops the stream if it was started."""

    if self._stream_started:
      ljm.eStreamStop(self._handle)

  def close(self) -> None:
    """Closes the Labjack if it was opened."""

    if self._handle is not None:
      ljm.close(self._handle)



# coding: utf-8

"""
Simulation of a temperature regulated furnace using a PID.

The temperature simulation is performed using a "Fake_motor". With the correct
parameters, this fake actuator can also represent the temperature in a furnace.

No hardware required.
"""

import crappy

SPEED = 20  # Speed multiplier of the simulation

P = .1
I = .1 * SPEED
D = 10 / SPEED


class Delay(crappy.modifier.Modifier):
  """Class to add a delay on the feedback."""

  def __init__(self, delay):
    super().__init__()
    self.delay = delay
    self.t = 't(s)'
    self.v = 'T'
    self.hist = []

  def evaluate(self, data):
    self.hist.append(data)
    r = dict(data)
    while self.hist and self.hist[0][self.t] + self.delay <= r[self.t]:
      del self.hist[0]
    v_table = [d[self.v] for d in self.hist]
    r[self.v] = sum(v_table) / len(v_table)
    return r


if __name__ == "__main__":
  g = crappy.blocks.Generator([
    dict(type='constant', condition="delay={}".format(300 / SPEED), value=200),
    # dict(type='constant', condition="delay=20", value=300),
    # dict(type='constant', condition="delay=20", value=400),
    dict(type='constant', condition=None, value=500)])

  furnace = crappy.blocks.Machine([dict(
    type='Fake_motor',
    cmd='pid',
    sim_speed=SPEED,
    mode='speed',
    speed_label='T',
    kv=1000,
    inertia=500,
    rv=.01,
    torque=-18,
    initial_speed=20,
    fv=1e-5
  )])

  pid = crappy.blocks.PID(P, I, D, input_label='T', out_max=1, out_min=0,
                          i_limit=(0.5, 0.5), send_terms=True)

  crappy.link(g, pid)
  crappy.link(pid, furnace)
  # Adding a delay on the feedback to account for the response time of the
  # furnace and sensor
  crappy.link(furnace, pid, modifier=Delay(20 / SPEED))

  graph = crappy.blocks.Grapher(('t(s)', 'T'))
  crappy.link(furnace, graph)

  graph_pid = crappy.blocks.Grapher(
      ('t(s)', 'p_term'),
      ('t(s)', 'i_term'),
      ('t(s)', 'd_term'),
      ('t(s)', 'pid'))
  crappy.link(pid, graph_pid)

  crappy.start()



# coding: utf-8

"""This file contains the code for binding the C++ Comedi library in Python.

The bindings haven't been tested for a while.
"""

from ctypes import *
from .._global import OptionalModule

try:
  comedi = cdll.LoadLibrary("/usr/lib/libcomedi.so")
  is_installed = True
except OSError:
  comedi = OptionalModule('libcomedi.so')
  is_installed = False


class comedi_t(Structure):
  pass


class comedi_range(Structure):
  pass


if is_installed:
  AREF_GROUND = 0
  lsampl_t = c_uint
  data = lsampl_t()  # To store data for data_read before returning it

  comedi_open = comedi.comedi_open
  comedi_open.restype = POINTER(comedi_t)  # Device handle
  comedi_open.argtypes = [c_char_p]  # Device path

  comedi_close = comedi.comedi_close
  comedi_close.restype = c_int  # Error code: 0 on success else -1
  comedi_close.argtypes = [POINTER(comedi_t)]  # device handle

  comedi_to_phys = comedi.comedi_to_phys
  comedi_to_phys.restype = c_double  # Physical value
  comedi_to_phys.argtypes = [lsampl_t, POINTER(comedi_range), lsampl_t]
  # Comedi value, range, maxdata

  comedi_from_phys = comedi.comedi_from_phys
  comedi_from_phys.restype = lsampl_t  # Comedi value
  comedi_from_phys.argtypes = [c_double, POINTER(comedi_range), lsampl_t]
  # Physical value, range, maxdata

  comedi_get_maxdata = comedi.comedi_get_maxdata
  comedi_get_maxdata.restype = lsampl_t  # Maxdata
  comedi_get_maxdata.argtypes = [POINTER(comedi_t), c_uint, c_uint]
  # handle, subdevice, channel

  comedi_get_range = comedi.comedi_get_range
  comedi_get_range.restype = POINTER(comedi_range)  # Range
  comedi_get_range.argtypes = [POINTER(comedi_t), c_uint, c_uint, c_uint]
  # handle, subdevice, channel, range_num

  comedi.comedi_data_read.restype = c_int  # Error code (1 if fine else -1)
  comedi.comedi_data_read.argtypes = [POINTER(comedi_t),
                                      c_uint, c_uint, c_uint,
                                      c_uint, POINTER(lsampl_t)]
  # Handle, subdevice, channel, range_num, aref, data pointer


  def comedi_data_read(*args):
    assert comedi.comedi_data_read(*(args +
                                     (byref(data),))) == 1, "Data read failed!"
    return data

  comedi_data_write = comedi.comedi_data_write
  comedi_data_write.restype = c_int  # Error code (1 if fine else -1)
  comedi_data_write.argtypes = [POINTER(comedi_t), c_uint, c_uint, c_uint,
                                c_uint, lsampl_t]
  # Handle, subdevice, channel, range_num, aref, data


else:
  AREF_GROUND = OptionalModule('libcomedi.so')
  lsampl_t = OptionalModule('libcomedi.so')
  data = OptionalModule('libcomedi.so')

  comedi_open = OptionalModule('libcomedi.so')

  comedi_close = OptionalModule('libcomedi.so')

  comedi_to_phys = OptionalModule('libcomedi.so')

  comedi_from_phys = OptionalModule('libcomedi.so')

  comedi_get_maxdata = OptionalModule('libcomedi.so')

  comedi_get_range = OptionalModule('libcomedi.so')

  comedi_data_read = OptionalModule('libcomedi.so')

  comedi_data_write = OptionalModule('libcomedi.so')



# coding: utf-8

# Todo:
#  Update with recent Python bindings

from time import time
from typing import Optional, Tuple
import numpy as np

from .camera import Camera
from .._global import OptionalModule
try:
  from . import clModule as Cl
except (ImportError, ModuleNotFoundError):
  Cl = OptionalModule("clModule", "CameraLink module was not compiled. Please "
                                  "make sure /opt/SiliconSoftware/xxx/lib64 "
                                  "exists and reinstall Crappy")


class Cl_camera(Camera):
  """Cameralink camera sensor."""

  def __init__(self) -> None:
    """Using the clModule, will open a cameraLink camera.

    Note:
      If a config file is specified, it will be used to configure the camera.

      If not set, it will be asked, unless set to :obj:`False` (or 0).

      Else, you must at least provide the camera type (eg: `"FullAreaGray8"`).

      Using a config file is recommended over changing all settings manually.
    """

    super().__init__()

    self.add_scale_setting("framespersec", 1, 200, self._get_framespersec,
                           self._set_framespersec)

  def open(self,
           numdevice: int = 0,
           config_file: Optional[str] = None,
           camera_type: Optional[str] = None,
           **kwargs) -> None:
    """Opens the camera."""

    if camera_type is None and config_file is not None:
      with open(config_file, 'r') as file:
        r = file.readlines()
      r = [s for s in r if s[:5] == "Typ='"]
      if r:
        camera_type = r[0][5:-3]

    if camera_type is None:
      raise AttributeError("No camera type or valid config file specified!")

    if 'format' in kwargs:
      f = kwargs['format']

    else:
      if camera_type[-1] == '8':
        f = Cl.FG_GRAY
      elif camera_type[-2:] == '16':
        f = Cl.FG_GRAY16
      elif camera_type[-2:] == '24':
        f = Cl.FG_COL24

      else:
        if config_file:
          with open(config_file, 'r') as file:
            r = file.readlines()
          r = [s for s in r if s[:10] == "FG_FORMAT="]
          if len(r) != 0:
            f = int(r[0].split('=')[1])
          else:
            raise ValueError("Could not determine the format")
        else:
          raise ValueError("Could not determine the format")

    self.cap = Cl.VideoCapture()
    self.cap.open(numdevice, camera_type, f)

    if config_file:
      self.cap.loadFile(config_file)

    self.set_all(**kwargs)

    self.cap.startAcq()
    self.cap.set(Cl.FG_TRIGGERMODE, 1)
    self.cap.set(Cl.FG_EXSYNCON, 1)

  def get_image(self) -> Tuple[float, np.ndarray]:
    """"""

    ret, frame = self.cap.read()
    t = time()
    if not ret:
      raise IOError("Could not read camera")

    return t, frame

  def close(self) -> None:
    """"""

    self.cap.stopAcq()
    self.cap.release()

  def _set_framespersec(self, val: float) -> None:
    self.cap.set(Cl.FG_FRAMESPERSEC, val)

  def _get_framespersec(self) -> float:
    return self.cap.get(Cl.FG_FRAMESPERSEC)

  def _set_h(self, val: int) -> None:
    self.cap.stopAcq()
    self.cap.set(Cl.FG_HEIGHT, val)
    self.cap.startAcq()

  def _set_w(self, val: int) -> None:
    self.cap.stopAcq()
    self.cap.set(Cl.FG_WIDTH, val)
    self.cap.startAcq()

  def _get_h(self) -> int:
    return self.cap.get(Cl.FG_HEIGHT)

  def _get_w(self):
    return self.cap.get(Cl.FG_WIDTH)



# coding: utf-8

# Parent class
from .camera import Camera, MetaCam
# Virtual cameras
from .fakeCamera import Fake_camera
from .file_reader import File_reader
# Physical cameras
from .webcam import Webcam
from .xiapi import Xiapi
from .pi_camera import Picamera
from .gstreamer import Camera_gstreamer
from .opencv import Camera_opencv
# Cameralink cameras
from .cameralink import Cl_camera
from .bispectral import Bispectral
from .jai import Jai, Jai8
from .seek_thermal_pro import Seek_thermal_pro

camera_list = MetaCam.classes



# coding: utf-8

from typing import Callable, Union, Optional, List, Dict, Any
import numpy as np
from pathlib import Path

from ..tool.videoextenso import LostSpotError, VideoExtenso as Ve
from ..tool.videoextensoConfig import VE_config
from .camera import Camera
from .displayer import Displayer


class Video_extenso(Camera):
  """This block can detect and track spots on images, and compute a strain
  value based on the displacement of the spots.

  It is meant to be used for following spots drawn on a sample during a tensile
  test, so that the local strain values can be deduced from the displacement of
  the spots. The spots are selected interactively in a GUI before the test
  starts.

  The timestamp, strain values, as well as the position of the detected spots
  are sent to downstream blocks for each received image.
  """

  def __init__(self,
               camera: str,
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               config: bool = True,
               display_images: bool = False,
               displayer_backend: Optional[str] = None,
               displayer_framerate: float = 5,
               verbose: bool = False,
               freq: float = 200,
               save_images: bool = False,
               img_name: str = "{self._n_loops:6d}_{t-self.t0:.6f}.tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               labels: List[str] = None,
               raise_on_lost_spot: bool = True,
               white_spots: bool = False,
               update_thresh: bool = False,
               num_spots: Optional[int] = None,
               safe_mode: bool = False,
               border: int = 5,
               min_area: int = 150,
               blur: int = 5,
               **kwargs) -> None:
    """Sets the args and initializes the camera.

    Args:
      camera: The name of the camera to control. See :ref:`Cameras` for an
        exhaustive list of available cameras.
      transform: A function taking an image as an argument and returning a
        transformed image. The original image is discarded and only the
        transformed one is kept for processing, display and saving.
      config: If :obj:`True`, a config window is shown before the test starts
        for interactively tuning the camera settings. It also allows selecting
        the spots to track.
      display_images: If :obj:`True`, a window displays the acquired images
        in low resolution during the test. This display is mainly intended for
        debugging and visual follow-up, but not for displaying high-quality
        images.
      displayer_backend: If ``display_images`` is :obj:`True`, the backend to
        use for the display window. Should be one of :
        ::

          'cv2', 'mpl'

        If not given, OpenCV will be used if available.
      displayer_framerate: If ``display_images`` is :obj:`True`, sets the
        maximum framerate for updating the display window. This setting allows
        limiting the resources used by the displayer. Note that the actual
        achieved framerate might differ, this is just the maximum limit.
      verbose: If :obj:`True`, the achieved framerate will be displayed in the
        console during the test.
      freq: If given, the block will try to loop at this frequency. If it is
        lower than the framerate of the camera, frames will be dropped. This
        argument can be used for limiting the achieved framerate when the
        camera doesn't support framerate control.
      save_images: If :obj:`True`, the acquired images are saved on the
        computer during the test. Note that saving images uses CPU, so the
        achieved performance might drop when this feature is in use.
      img_name: If ``save_images`` is :obj:`True`, the template for naming the
        recorded images. It is evaluated as an `f-string`, and must contain the
        file extension at the end. For building the `f-string`, the
        ``self._n_loops`` attribute holds the loop number, and ``t-self.t0``
        holds the current timestamp.
      save_folder: If ``save_images`` is :obj:`True`, the directory to save
        images to. If it doesn't exist, it will be created. If not given, the
        images are saved in a folder named `Crappy_images` and created next to
        the file being run.
      save_period: If ``save_images`` is :obj:`True`, only one every this
        number of images will be saved.
      save_backend: The backend to use for saving the images. Should be one
        of :
        ::

          'sitk', 'pil', 'cv2'

        If not specified, SimpleITK will be used if available, then OpenCV as a
        second choice, and finally Pillow if none of the others was available.
      image_generator: A function taking two floats as arguments, and returning
        an image. It is only used for demonstration without camera in the
        examples, and isn't meant to be used in an actual test.
      labels: A :obj:`list` containing the labels to send to downstream blocks,
        carrying the information about the position of the tracked spots and
        the strain values. If not given, the default labels are :
        ::

          ['t(s)', 'meta', 'Coord(px)', 'Eyy(%)', 'Exx(%)']

      raise_on_lost_spot: If :obj:`True`, an exception is raised as soon as
        the block is losing track of a spot, what causes the test to stop.
        Otherwise, the block simply stops processing incoming images but
        doesn't raise any exception.
      white_spots: If :obj:`True`, detects white objects on a black background,
        else black objects on a white background.
      update_thresh: If :obj:`True`, the threshold for detecting the spots is
        re-calculated for each new image. Otherwise, the first calculated
        threshold is kept for the entire test. The spots are less likely to be
        lost with adaptive threshold, but the measurement will be more noisy.
        Adaptive threshold may also yield inconsistent results when spots are
        lost.
      num_spots: The number of spots to detect, between 1 and 4. The class will
        then try to detect this exact number of spots, and won't work if not
        enough spots can be found. If this argument is not given, at most 4
        spots can be detected but the class will work with any number of
        detected spots between 1 and 4.
      safe_mode: If :obj:`True`, the class will stop and raise an exception as
        soon as overlapping is detected. Otherwise, it will first try to reduce
        the detection window to get rid of overlapping. This argument should be
        used when inconsistency in the results may have critical consequences.
      border: When searching for the new position of a spot, the class will
        search in the last known bounding box of this spot plus a few
        additional pixels in each direction. This argument sets the number of
        additional pixels to use. It should be greater than the expected
        "speed" of the spots, in pixels / frame. But if it's too big, noise or
        other spots might hinder the detection.
      min_area: The minimum area an object should have to be potentially
        detected as a spot. The value is given in pixels, as a surface unit.
        It must of course be adapted depending on the resolution of camera and
        the size of the spots to detect.
      blur: The size in pixels of the kernel to use for applying a median blur
        to the image before the spot detection. If not given, no blurring is
        performed. A slight blur improves the spot detection by smoothening the
        noise, but also takes a bit more time compared to no blurring.
      **kwargs: Any additional argument to pass to the camera.
    """

    super().__init__(camera=camera,
                     transform=transform,
                     config=False,
                     display_images=False,
                     displayer_backend=displayer_backend,
                     displayer_framerate=displayer_framerate,
                     software_trig_label=None,
                     verbose=verbose,
                     freq=freq,
                     save_images=save_images,
                     img_name=img_name,
                     save_folder=save_folder,
                     save_period=save_period,
                     save_backend=save_backend,
                     image_generator=image_generator,
                     **kwargs)

    self._ve = Ve(white_spots=white_spots,
                  update_thresh=update_thresh,
                  num_spots=num_spots,
                  safe_mode=safe_mode,
                  border=border,
                  min_area=min_area,
                  blur=blur)

    self.labels = ['t(s)', 'meta', 'Coord(px)',
                   'Eyy(%)', 'Exx(%)'] if labels is None else labels

    # Setting the args
    self._display_images = display_images
    self._raise_on_lost_spot = raise_on_lost_spot
    self._config_ve = config
    self._lost_spots = False

    # Instantiating the displayer
    if display_images:
      self._displayer_ve = Displayer(f"Displayer {camera} "
                                     f"{Camera.cam_count[self._camera_name]}",
                                     displayer_framerate,
                                     displayer_backend)
    else:
      self._displayer_ve = None

  def prepare(self) -> None:
    """Opening the camera, starting the Video Extenso config and the tracker
    processes."""

    super().prepare()

    if self._config_ve:
      config = VE_config(self._camera, self._ve)
      config.main()

    self._ve.start_tracking()

    if self._displayer_ve is not None:
      self._displayer_ve.prepare()

  def finish(self) -> None:
    """Stopping the tracker processes, the displayer and the camera."""

    self._ve.stop_tracking()

    if self._displayer_ve is not None:
      self._displayer_ve.finish()

    super().finish()

  def _additional_loop(self, meta: Dict[str, Any], img: np.ndarray) -> None:
    """Gets the data from the latest image, sends it, and updates the
    display."""

    # Trying to process the incoming images
    if not self._lost_spots:
      try:
        data = self._ve.get_data(img)
        if data is not None:
          self.send([meta['t(s)'], meta, *data])
      except LostSpotError:
        self._ve.stop_tracking()
        # Raising if specified by the user
        if self._raise_on_lost_spot:
          raise
        # Otherwise, simply setting a flag so that no additional processing is
        # performed
        else:
          self._lost_spots = True
          print("[VideoExtenso] Spots lost, not processing data anymore !")

    # Updating the display
    if self._displayer_ve is not None:

      # Adding the contours on the image only if the spots were not lost
      if not self._lost_spots:
        for spot in self._ve.spots:
          self._draw_box(img, spot)

      # Updating the display in all cases
      self._displayer_ve.update(img)



# coding: utf-8

import numpy as np
from typing import Optional
from time import time
from .._global import OptionalModule

try:
  from PIL import ImageTk, Image
except (ModuleNotFoundError, ImportError):
  ImageTk = OptionalModule("Pillow")
  Image = OptionalModule("Pillow")

try:
  import matplotlib.pyplot as plt
except (ModuleNotFoundError, ImportError):
  plt = OptionalModule("matplotlib")

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class Displayer:
  """This class manages the display of images captured by a camera object.

  It can either work with Matplotlib or OpenCV as a backend, OpenCV being the
  fastest. It tries to display images at a given framerate, dropping part of
  the received frames if necessary.

  The images are displayed with a maximum resolution of 640x480, and are
  resized to match that resolution if necessary. Similarly, the maximum bit
  depth is 8 bits, and the images are cast if necessary. Resizing and casting
  are anyway less demanding on the CPU than displaying big images.
  """

  def __init__(self,
               title: str,
               framerate: float,
               backend: Optional[str] = None) -> None:
    """Sets the args and the other instance attributes, and looks for an
    available display backend if none was specified.

    Args:
      title: The title to display on the display window.
      framerate: The maximum framerate for displaying the images. To achieve
        this framerate, part of the received frames are simply dropped.
      backend: The backend to use for displaying the images. Should be one of:
        ::

          'cv2', 'mpl'
    """

    self._title = title
    self._framerate = framerate

    # Selecting the backend if no backend was specified
    if backend is None:
      if not isinstance(cv2, OptionalModule):
        self._backend = 'cv2'
      elif not isinstance(plt, OptionalModule):
        self._backend = 'mpl'
      else:
        raise ModuleNotFoundError("Neither opencv-python nor matplotlib could "
                                  "be imported, no backend found for "
                                  "displaying the images")

    elif backend in ('cv2', 'mpl'):
      self._backend = backend
    else:
      raise ValueError("The backend argument should be either 'cv2' or "
                       "'mpl' !")

    # Setting other instance attributes
    self._ax = None
    self._fig = None
    self._last_upd = time()

  def prepare(self) -> None:
    """Calls the right prepare method depending on the chosen backend."""

    if self._backend == 'cv2':
      self._prepare_cv2()
    elif self._backend == 'mpl':
      self._prepare_mpl()

  def update(self, img: np.ndarray) -> None:
    """Ensures the target framerate is respected, and calls the right update
    method depending on the chosen backend.

    Args:
      img: The image to display on the displayer.
    """

    # Making sure the image is not being refreshed too often
    if time() - self._last_upd < 1 / self._framerate:
      return
    else:
      self._last_upd = time()

    # Casts the image to uint8 if it's not already in this format
    if img.dtype != np.uint8:
      if np.max(img) > 255:
        factor = min((i for i in range(1, 10) if np.max(img) / 2 ** i < 256))
        img = (img / 2 ** factor).astype(np.uint8)
      else:
        img = img.astype(np.uint8)

    # Calling the right prepare method
    if self._backend == 'cv2':
      self._update_cv2(img)
    elif self._backend == 'mpl':
      self._update_mpl(img)

  def finish(self) -> None:
    """Calls the right finish method depending on the chosen backend."""

    if self._backend == 'cv2':
      self._finish_cv2()
    elif self._backend == 'mpl':
      self._finish_mpl()

  def _prepare_cv2(self) -> None:
    """Instantiates the display window of cv2."""

    try:
      flags = cv2.WINDOW_NORMAL | cv2.WINDOW_KEEPRATIO
    except AttributeError:
      flags = cv2.WINDOW_NORMAL
    cv2.namedWindow(self._title, flags)

  def _prepare_mpl(self) -> None:
    """Creates a Matplotlib figure."""

    plt.ion()
    self._fig, self._ax = plt.subplots()

  def _update_cv2(self, img: np.ndarray) -> None:
    """Reshapes the image to a maximum shape of 640x480 and displays it."""

    if img.shape[0] > 480 or img.shape[1] > 640:
      factor = min(480 / img.shape[0], 640 / img.shape[1])
      img = cv2.resize(img, (int(img.shape[1] * factor),
                             int(img.shape[0] * factor)))

    cv2.imshow(self._title, img)
    cv2.waitKey(1)

  def _update_mpl(self, img: np.ndarray) -> None:
    """Reshapes the image to a dimension inferior or equal to 640x480 and
    displays it."""

    if img.shape[0] > 480 or img.shape[1] > 640:
      factor = min((i for i in range(2, 10) if img.shape[0] / i <= 480
                    and img.shape[1] / i <= 640))
      img = img[::factor, ::factor]

    self._ax.clear()
    self._ax.imshow(img, cmap='gray')
    plt.pause(0.001)
    plt.show()

  def _finish_cv2(self) -> None:
    """Destroys the opened cv2 window."""

    cv2.destroyWindow(self._title)

  def _finish_mpl(self) -> None:
    """Destroys the opened Matplotlib window."""

    plt.close(self._fig)



# coding: utf-8

from time import sleep
from re import findall
from .actuator import Actuator
from .._global import OptionalModule

try:
  from serial import Serial
except (ModuleNotFoundError, ImportError):
  Serial = OptionalModule('pyserial')


class Tra6ppd(Actuator):
  """Drives the TRA6PPD linear actuator in position.

  Warning:
    This actuator cannot handle a high serial messages rate. It is recommended
    to set the frequency of the corresponding :ref:`Machine` block to a few
    dozen Hz at most.

  Note:
    This actuator ignores new position commands while it is moving.
  """

  def __init__(self,
               baudrate: int = 57600,
               port: str = '/dev/ttyUSB0') -> None:
    """Sets the instance attributes.

    Args:
      baudrate: The serial baud rate.
      port: Path to the serial port to connect to.
    """
    self._max_speed = 0.2
    self._min_position = 0
    self._max_position = 6
    self._ser = Serial(port, baudrate=baudrate, timeout=0.1)

  def open(self) -> None:
    """Resets the device and performs homing."""

    # First, reset the device
    self._ser.write(b'1RS\r\n')
    sleep(5)

    # Then, perform homing (may take up to 45s)
    self._ser.write(b'1OR\r\n')
    for i in range(5, 0, -1):
      print(f"[TRA6PPD] Performing homing, {10 * i} seconds left.")
      sleep(10)

  def close(self) -> None:
    """Just closes the serial port."""

    self._ser.close()

  def stop(self) -> None:
    """Stops the motor and sets the device to "disable" state."""

    self._ser.write(b'ST\r\n')
    self._ser.write(b'1MMO\r\n')

  def set_position(self,
                   position: float,
                   speed: float = 0.2) -> None:
    """Sends the actuator a command to reach a given position.

    The command is ignored if the actuator is already moving.

    Args:
      position: The position to reach. Should be between 0 and 6 mm.
      speed: The speed at which the actuator should move. Should be between 0
        and 0.2 mm/s.
    """

    # Clamping the speed command into the allowed interval
    speed_clamped = max(min(self._max_speed, speed), 0)

    # Sending the speed command
    speed_cmd = f'1VA{speed_clamped:.5f}\r\n'
    self._ser.write(speed_cmd.encode())

    # Ignoring the position command if it is out of range
    if not self._min_position <= position <= self._max_position:
      print(f"[TRA6PPD] WARNING : The requested position {position} is out of "
            f"range ! Ignoring.")
    else:
      # Sending the position command
      pos_command = f'1PA{position:.5f}\r\n'
      self._ser.write(pos_command.encode())

  def get_position(self) -> float:
    """Reads the current position.

    Returns:
      Current position of the motor.
    """

    # Sending the read command
    self._ser.write(b'1TP?\r\n')
    # Using a regular expression to parse the answer
    ans = self._ser.readline().decode()
    actuator_pos = float(findall(r'\d\.\d+', ans)[0])
    return actuator_pos



# coding: utf-8

from time import time
from typing import Union, Optional, List
from .inout import InOut
from .._global import OptionalModule
from ..tool import ft232h_server as ft232h, Usb_server, i2c_msg_ft232h

try:
  import adafruit_mprls
except (ImportError, ModuleNotFoundError):
  adafruit_mprls = OptionalModule('adafruit_mprls')

try:
  import board
except (ImportError, ModuleNotFoundError):
  board = OptionalModule('board', 'Blinka is necessary to use the I2C bus')

try:
  from digitalio import DigitalInOut
except (ImportError, ModuleNotFoundError):
  DigitalInOut = OptionalModule('digitalio',
                                'Blinka is necessary to access the GPIOs')

try:
  from smbus2 import SMBus, i2c_msg
except (ImportError, ModuleNotFoundError):
  SMBus = i2c_msg = OptionalModule('smbus2')

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")

mprls_status_bits = {'busy': 0x20,
                     'memory error': 0x04,
                     'math saturation': 0x01}
mprls_backends = ['Pi4', 'blinka', 'ft232h']


class Mprls(Usb_server, InOut):
  """The Mprls inout is meant for reading pressure from Adafruit's Mprls
    pressure sensor.

    It communicates over I2C with the sensor.
    """

  def __init__(self,
               backend: str,
               eoc_pin: Optional[Union[str, int]] = None,
               device_address: int = 0x18,
               i2c_port: int = 1,
               ft232h_ser_num: Optional[str] = None) -> None:
    """Initializes the parent class and opens the I2C bus.

    Args:
      backend (:obj:`str`): Should be one of :
        ::

          'Pi4', 'blinka', 'ft232h'

        The `'Pi4'` backend is optimized but only works on boards supporting
        the :mod:`smbus2` module, like the Raspberry Pis. The `'blinka'`
        backend may be less performant and requires installing Adafruit's
        modules, but these modules are compatible with and maintained on a wide
        variety of boards. The `'ft232h'` backend allows controlling the
        MPRLS from a PC using Adafruit's FT232H USB to I2C adapter. See
        :ref:`Crappy for embedded hardware` for details.
      eoc_pin (:obj:`int` or :obj:`str`, optional): Optionally, reads the end
        of conversion signal from a GPIO rather than from an I2C message.
        Speeds up the reading and decreases the traffic on the bus, but
        requires one extra wire. With the backend `'Pi4'`, give the index of
        the GPIO in BCM convention. With the `'ft232h'` backend, give the name
        of the GPIO in the format `Dx` or `Cx`. With the backend `'blinka'`,
        it should be a string but the syntax varies according to the board.
        Refer to blinka's documentation for more information.
      device_address (:obj:`int`, optional): The I2C address of the MPRLS.
        The address of the devices sold by Adafruit is `0x18`, but other
        suppliers may sell it with another address.
      i2c_port (:obj:`int`, optional): The I2C port over which the MPRLS
        should communicate. On most Raspberry Pi models the default I2C port is
        `1`.
      ft232h_ser_num (:obj:`str`, optional): If backend is `'ft232h'`, the
        serial number of the FT232H to use for communication.
    """

    if not isinstance(backend, str) or backend not in mprls_backends:
      raise ValueError("backend should be in {}".format(mprls_backends))
    self._backend = backend

    Usb_server.__init__(self,
                        serial_nr=ft232h_ser_num if ft232h_ser_num else '',
                        backend=backend)
    current_file, block_number, command_file, answer_file, block_lock, \
        current_lock = super().start_server()

    InOut.__init__(self)

    if backend == 'ft232h':
      self._bus = ft232h(mode='I2C',
                         block_number=block_number,
                         current_file=current_file,
                         command_file=command_file,
                         answer_file=answer_file,
                         block_lock=block_lock,
                         current_lock=current_lock,
                         serial_nr=ft232h_ser_num)

    if not isinstance(device_address, int):
      raise TypeError("device_address should be an integer.")
    self._address = device_address

    if not isinstance(i2c_port, int):
      raise TypeError("i2c_port should be an integer.")
    self._i2c_port = i2c_port

    if eoc_pin is not None:
      if backend == 'blinka' and not isinstance(eoc_pin, str):
        raise TypeError('eoc_pin should be a string when using the blinka '
                        'backend !')
      elif backend == 'ft232h' and not isinstance(eoc_pin, str):
        raise TypeError('eoc_pin should be a string when using the ft232h '
                        'backend !')
      elif backend == 'Pi4' and not isinstance(eoc_pin, int):
        raise TypeError('eoc_pin should be an int when using the Pi4 '
                        'backend !')
    self._eoc_pin = eoc_pin

  def open(self) -> None:
    """Opens the I2C bus."""

    if self._backend == 'Pi4':
      self._bus = SMBus(self._i2c_port)
      if self._eoc_pin is not None:
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self._eoc_pin, GPIO.IN)

    elif self._backend == 'blinka':
      if self._eoc_pin is not None:
        eoc = DigitalInOut(getattr(board, self._eoc_pin))
      else:
        eoc = None
      self._mpr = adafruit_mprls.MPRLS(board.I2C(),
                                       psi_min=0,
                                       psi_max=25,
                                       eoc_pin=eoc)

    self._i2c_msg = i2c_msg_ft232h if self._backend == 'ft232h' else i2c_msg

  def get_data(self) -> List[float]:
    """Reads the pressure value.

    Returns:
      The timestamp and the pressure value in hPa.
    """

    if self._backend == 'blinka':
      pres = self._mpr.pressure

    else:
      # Starting conversion
      self._bus.i2c_rdwr(self._i2c_msg.write(self._address,
                                             [0xAA, 0x00, 0x00]))
      # Waiting for conversion to complete
      t0 = time()
      while not self._data_available():
        if time() - t0 > 0.1:
          raise TimeoutError('Waited too long for data to be ready')

      # Reading conversion result
      read = self._i2c_msg.read(self._address, 4)
      self._bus.i2c_rdwr(read)
      out = list(read)
      # Checking if anu error occurred
      if out[0] & mprls_status_bits['memory error']:
        raise RuntimeError("A memory error occurred on the MPRLS")
      elif out[0] & mprls_status_bits['math saturation']:
        raise RuntimeError("A math saturation error occurred on the MPRLS")
      # Extracting conversion result as an integer
      ret = (out[1] << 16) | (out[2] << 8) | out[3]
      # Converting to hPa
      pres = 68.947572932 * (ret - 0x19999A) * 25 / (0xE66666 - 0x19999A)

    return [time(), pres]

  def close(self) -> None:
    """Closes the I2C bus."""

    if self._backend != 'blinka':
      self._bus.close()

    if self._backend == 'Pi4' and self._eoc_pin is not None:
      GPIO.cleanup()

  def _data_available(self) -> bool:
    """Returns :obj:`True` if data is available, :obj:`False` otherwise."""

    # EOC signal from the I2C communication
    if self._eoc_pin is None:
      wait = self._i2c_msg.read(self._address, 1)
      self._bus.i2c_rdwr(wait)
      return not list(wait)[0] & mprls_status_bits['busy']
    # EOC signal from a GPIO
    elif self._backend == 'ft232h':
      return bool(self._bus.get_gpio(self._eoc_pin))
    else:
      return bool(GPIO.input(self._eoc_pin))



# coding: utf-8

import numpy as np
from typing import Callable, Union, Optional, List, Dict, Any
from pathlib import Path

from ..tool import GPUCorrel as GPUCorrel_tool
from .camera import Camera
from .displayer import Displayer


class GPUCorrel(Camera):
  """This blocks projects a displacement field on a given base of fields, and
  sends the decomposition to downstream blocks.

  It relies on the :ref:`GPU Correl` class. The displacement is calculated for
  the entire image, and it is not possible to select a region of interest.
  """

  def __init__(self,
               camera: str,
               fields: List[str],
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               display_images: bool = False,
               displayer_backend: Optional[str] = None,
               displayer_framerate: float = 5,
               verbose: int = 0,
               freq: float = 200,
               save_images: bool = False,
               img_name: str = "{self._n_loops:6d}_{t-self.t0:.6f}.tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               labels: Optional[List[str]] = None,
               discard_limit: float = 3,
               discard_ref: int = 5,
               img_ref: Optional[np.ndarray] = None,
               levels: int = 5,
               resampling_factor: float = 2,
               kernel_file: Optional[Union[str, Path]] = None,
               iterations: int = 4,
               mask: Optional[np.ndarray] = None,
               mul: float = 3,
               res: bool = False,
               **kwargs) -> None:
    """Sets the args and initializes the camera object.

    Args:
      camera: The name of the camera to control. See :ref:`Cameras` for an
        exhaustive list of available cameras.
      fields: A :obj:`list` of :obj:`str` representing the base of fields on
        which the image will be projected during correlation. The possible
        fields are :
        ::

          'x', 'y', 'r', 'exx', 'eyy', 'exy', 'eyx', 'exy2', 'z'

      transform: A function taking an image as an argument and returning a
        transformed image. The original image is discarded and only the
        transformed one is kept for processing, display and saving.
      display_images: If :obj:`True`, the difference between the original and
        the displaced image after correlation will be displayed. `128` means
        no difference, lighter means positive and darker negative.
      displayer_backend: If ``display_images`` is :obj:`True`, the backend to
        use for the display window. Should be one of :
        ::

          'cv2', 'mpl'

        If not given, OpenCV will be used if available.
      displayer_framerate: If ``display_images`` is :obj:`True`, sets the
        maximum framerate for updating the display window. This setting allows
        limiting the resources used by the displayer. Note that the actual
        achieved framerate might differ, this is just the maximum limit.
      verbose: The verbose level as an integer, between `0` and `3`. At level
        `0` no information is printed, and at level `3` so much information is
        printed that is slows the code down.
      freq: If given, the block will try to loop at this frequency. If it is
        lower than the framerate of the camera, frames will be dropped. This
        argument can be used for limiting the achieved framerate when the
        camera doesn't support framerate control.
      save_images: If :obj:`True`, the acquired images are saved on the
        computer during the test. Note that saving images uses CPU, so the
        achieved performance might drop when this feature is in use.
      img_name: If ``save_images`` is :obj:`True`, the template for naming the
        recorded images. It is evaluated as an `f-string`, and must contain the
        file extension at the end. For building the `f-string`, the
        ``self._n_loops`` attribute holds the loop number, and ``t-self.t0``
        holds the current timestamp.
      save_folder: If ``save_images`` is :obj:`True`, the directory to save
        images to. If it doesn't exist, it will be created. If not given, the
        images are saved in a folder named `Crappy_images` and created next to
        the file being run.
      save_period: If ``save_images`` is :obj:`True`, only one every this
        number of images will be saved.
      save_backend: The backend to use for saving the images. Should be one
        of :
        ::

          'sitk', 'pil', 'cv2'

        If not specified, SimpleITK will be used if available, then OpenCV as a
        second choice, and finally Pillow if none of the others was available.
      image_generator: A function taking two floats as arguments, and returning
        an image. It is only used for demonstration without camera in the
        examples, and isn't meant to be used in an actual test.
      labels: A :obj:`list` containing the labels to send to downstream blocks,
        carrying the displacement projected on the given basis of fields. If
        not given, the labels list is just a copy of the fields list, with
        ``'t(s)'`` added in position 0, ``'meta'`` in position 1, and ``'res'``
        added in last position if the ``res`` argument is :obj:`True`.
      discard_limit: If given, the data is sent to downstream blocks only if
        the residuals are lower than the average of the last few residuals
        multiplied by this value.
      discard_ref: When checking whether the data should be sent based on the
        ``discard_limit`` argument, only that many previous values will be
        considered when calculating the average of the last residuals.
      img_ref: The reference image to which all the acquired images will be
        compared for performing the correlation. If not given, the first
        acquired images will be used as the reference image.
      levels: Number of levels of the pyramid. More levels may help converging
        on images with large strain, but may fail on images that don't contain
        low spatial frequency. Fewer levels mean that the program runs faster.
      resampling_factor: the factor by which the resolution is divided between
        each stage of the pyramid. A low factor ensures coherence between the
        stages, but is more computationally intensive. A high factor allows
        reaching a finer detail level, but may lead to a coherence loss between
        the stages.
      kernel_file: The path to the file containing the kernels to use for the
        correlation. Can be a :obj:`pathlib.Path` object or a :obj:`str`.
      iterations: The maximum number of iterations to run before returning the
        results. The results may be returned before if the residuals start
        increasing.
      mask: The mask used for weighting the region of interest on the image. It
        is generally used to prevent unexpected behavior on the border of the
        image.
      mul: The scalar by which the direction will be multiplied before being
        added to the solution. If it's too high, the convergence will be fast
        but there's a risk that to go past the solution and to diverge. If it's
        too low, the convergence will be slower and require more iterations.
        `3` was found to be an acceptable value in most cases, but it is
        recommended to tune this value for each application so that the
        convergence is neither too slow nor too fast.
      res: If :obj:`True`, the residuals will be sent to downstream blocks
        along with the other information under the label ``'res'``.
      **kwargs: Any additional argument to pass to the camera.
    """

    super().__init__(camera=camera,
                     transform=transform,
                     config=False,
                     display_images=False,
                     displayer_backend=None,
                     displayer_framerate=5,
                     software_trig_label=None,
                     verbose=bool(verbose),
                     freq=freq,
                     save_images=save_images,
                     img_name=img_name,
                     save_folder=save_folder,
                     save_period=save_period,
                     save_backend=save_backend,
                     image_generator=image_generator,
                     **kwargs)

    self._correl = GPUCorrel_tool(context=None,
                                  verbose=verbose,
                                  levels=levels,
                                  resampling_factor=resampling_factor,
                                  kernel_file=kernel_file,
                                  iterations=iterations,
                                  mask=mask,
                                  ref_img=img_ref,
                                  mul=mul,
                                  fields=fields)

    # Setting the labels
    if labels is None:
      self.labels = ['t(s)', 'meta'] + fields
    else:
      self.labels = labels

    if res:
      self.labels.append('res')
    self._calc_res = res

    if 2 + len(fields) + int(res) != len(self.labels):
      raise ValueError("The number of fields is inconsistent with the number "
                       "of labels !\nMake sure that the time label was given")

    # Setting the args
    self._discard_limit = discard_limit
    self._discard_ref = discard_ref
    self._img_ref = img_ref
    self._res_hist = [np.inf]

    # Initializing the displayer
    if display_images:
      self._displayer_gpu = Displayer(f"Displayer {camera} "
                                      f"{Camera.cam_count[self._camera_name]}",
                                      displayer_framerate,
                                      displayer_backend)
    else:
      self._displayer_gpu = None

  def prepare(self) -> None:
    """Opens the camera, prepares the displayer and sets the reference image if
    one was given."""

    super().prepare()

    if self._img_ref is not None:
      self._correl.set_img_size(self._img_ref.shape)
      self._correl.prepare()

    if self._displayer_gpu is not None:
      self._displayer_gpu.prepare()

  def begin(self) -> None:
    """Acquires an image and sets it as the reference image if no image was
    given as reference in the arguments."""

    if self._img_ref is None:
      _, img = self._camera.get_image()

      self._correl.set_img_size(img.shape)
      self._correl.set_orig(img.astype(np.float32))
      self._correl.prepare()

  def finish(self) -> None:
    """Closes the displayer and the camera object."""

    if self._displayer_gpu is not None:
      self._displayer_gpu.finish()

    self._correl.clean()
    super().finish()

  def _additional_loop(self, meta: Dict[str, Any], img: np.ndarray) -> None:
    """Gets the updated fields of displacement, and sends them to downstream
    blocks.

    Optionally, also checks if the residuals are low enough for safely sending
    the data.
    """

    out = [meta['t(s)'], meta]
    out += self._correl.get_disp(img.astype(np.float32)).tolist()

    if self._calc_res:
      res = self._correl.get_res()
      out.append(res)

      if self._discard_limit:
        self._res_hist.append(res)
        self._res_hist = self._res_hist[-self._discard_ref - 1:]

        if res > self._discard_limit * np.average(self._res_hist[:-1]):
          print("[GPUCorrel] Residual too high, not sending values")
          return

    self.send(out)



# coding: utf-8

from typing import Dict, Any, List, Union, Tuple

from .modifier import Modifier


class Offset(Modifier):
  """This Modifier offsets every value of the given labels by a constant. This
  constant is calculated so that for each label the first returned value is
  equal to a user-defined target.

  For example if for a given label the target is `6` and the first received
  value is `3`, the Modifier will add `3` to each value received over this
  label.

  This Modifier can be used for example when measuring a variable that should
  start at `0` (like a force) but doesn't because of a sensor offset. It can
  also just be used to plot nicer figures. It is not very accurate as it is
  only based on a single data point for the offset calculation. The
  ``make_zero`` argument of the :ref:`IOBlock` is a better alternative if
  precision is required when offsetting the data.
  """

  def __init__(self,
               labels: Union[str, List[str], Tuple[str, ...]],
               offsets: Union[float, List[float], Tuple[float, ...]]) -> None:
    """Sets the args and initializes the parent class.

    Args:
      labels: The labels to offset. Can be given as a single label, a
        :obj:`list` of labels or a :obj:`tuple` of labels.
      offsets: For each label, the target for the first received value. Can be
        given as a single value, a :obj:`list` of values or a :obj:`tuple` of
        values.
    """

    super().__init__()

    # Handling the case when only one label needs to be offset
    if not isinstance(labels, list) and not isinstance(labels, tuple):
      labels = (labels,)
    if not isinstance(offsets, list) and not isinstance(offsets, tuple):
      offsets = (offsets,)

    # Checking that the number of offsets match the number of labels
    if len(offsets) != len(labels):
      raise ValueError("As many offsets as there are labels should be given.")

    # Associating each offset to its label
    self._offsets = {label: offset for label, offset in zip(labels, offsets)}

    self._compensations = None
    self._compensated = False

  def evaluate(self, data: Dict[str, Any]) -> Dict[str, Any]:
    """If the compensations are not set, sets them, anf then offsets the
    required labels."""

    # During the first loop, calculating the compensation values
    if not self._compensated:
      self._compensations = {label: -data[label] + offset for label, offset in
                             self._offsets.items()}
      self._compensated = True

    # Compensating the data to match the target offset value
    for label in self._offsets:
      data[label] += self._compensations[label]

    return data



# coding: utf-8

from .cameraConfig import Camera_config
from .disveConfig import DISVE_config
from .videoextensoConfig import VE_config
from .gpucorrel import GPUCorrel
from .discorrel import DISCorrel
from .discorrelConfig import DISConfig
from .disve import DISVE
from .ft232h import ft232h, i2c_msg_ft232h, ft232h_pin_nr, ft232h_server
from .ft232h import Usb_server
from .cameraConfigTools import Box, Spot_boxes



# coding: utf-8

import numpy as np
from typing import Dict, Any, Union, List, Tuple

from .modifier import Modifier


class Demux(Modifier):
  """Modifier converting a stream into a regular Crappy :obj:`dict` giving for
  each label a single value.

  The single value is either the first value of a column/row, or the average
  of the row/column values. This Modifier is mainly meant for linking streaming
  :ref:`IOBlock` blocks to :ref:`Grapher` blocks, as it is otherwise impossible
  to plot their data.
  """

  def __init__(self,
               labels: Union[str, List[str], Tuple[str, ...]],
               stream_label: str = "stream",
               mean: bool = False,
               time_label: str = "t(s)",
               transpose: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      labels: The labels corresponding to the rows or columns of the stream.
        It can be either a single label, or a :obj:`list` of labels, or a
        :obj:`tuple` of labels. They must be given in the same order as they
        appear in the stream. If fewer labels are given than there are rows or
        columns in the stream, only the data from the first rows/columns will
        be retrieved.
      stream_label: The label carrying the stream.
      mean: If :obj:`True`, the returned value will be the average of the
        stream data. Otherwise, it will be the first value.
      time_label: The label carrying the time information.
      transpose: If :obj:`True`, each label corresponds to a row in the stream.
        Otherwise, a label corresponds to a column in the stream.
    """

    super().__init__()

    if isinstance(labels, list) or isinstance(labels, tuple):
      self._labels = labels
    else:
      self._labels = (labels,)
    self._stream_label = stream_label
    self._mean = mean
    self._time_label = time_label
    self._transpose = transpose

  def evaluate(self, data: Dict[str, np.ndarray]) -> Dict[str, Any]:
    """Retrieves for each label its value in the stream, also gets the
    corresponding timestamp, and returns them."""

    # If there are no rows or no column, cannot perform the demux
    if 0 in data[self._stream_label].shape:
      return data

    # Getting either the average or the first value for each label
    for i, label in enumerate(self._labels):
      # The data of a given label is on a same row
      if self._transpose:
        if self._mean:
          data[label] = np.mean(data[self._stream_label][i, :])
        else:
          data[label] = data[self._stream_label][i, 0]
      # The data of a given label is on a same column
      else:
        if self._mean:
          data[label] = np.mean(data[self._stream_label][:, i])
        else:
          data[label] = data[self._stream_label][0, i]

    # Discarding the raw data
    del data[self._stream_label]

    # Keeping either the average or the first time value
    if self._mean:
      data[self._time_label] = np.mean(data[self._time_label])
    else:
      data[self._time_label] = np.squeeze(data[self._time_label])[0]

    return data



# coding: utf-8

from time import time
from struct import pack, unpack
from typing import List

from .inout import InOut
from .._global import OptionalModule
try:
  from pymodbus.client.tcp import ModbusTcpClient
except (ModuleNotFoundError, ImportError):
  ModbusTcpClient = OptionalModule("pymodbus", "Cannot use KollMorgenVariator")


coil_addr = {'power': 0,
             'move_abs': 1,
             'move_rel': 2,
             'move_vel': 3,
             'stop': 4,
             'ack_error': 5}

reg_addr = {'position': 0,
            'distance': 2,
            'velocity': 4,
            'acc': 5,
            'dec': 6,
            'fstdec': 7,
            'direction': 8}

input_reg_addr = {'act_speed': 0,
                  'act_position': 2,
                  'axis_state': 4}


class Koll(InOut):
  """This class can communicate with a KollMorgen AKD PDMM programmable
   multi-axis controller.

   It can either drive it in speed or in position. Multiple axes can be driven.
   The values of the current speeds or positions can also be retrieved.
   """

  def __init__(self,
               axes: List[int],
               mode: str = 'position',
               host: str = '192.168.0.109',
               port: int = 502) -> None:
    """Sets the args and initializes the parent class.

    Args:
      axes: A :obj:`list` containing the motors/axes to drive, given as
        :obj:`int`.
      mode: Should be either `'speed'` or `'position'`. Whether the axes
        should be driven in speed or in position.
      host: The IP address of the variator, given as a :obj:`str`.
      port: The network port over which to communicate with the variator, as
        an :obj:`int`.
    """

    super().__init__()

    # Making sure the given mode is correct
    if mode not in ('speed', 'position'):
      raise ValueError("[KollMorgen] the mode argument should be either "
                       "'speed' or 'position' !")

    self._axes = axes
    self._mode = mode

    self._variator = ModbusTcpClient(host=host, port=port)

  def open(self) -> None:
    """Connects to the variator over modbus."""

    self._variator.connect()

  def get_data(self) -> List[float]:
    """For each motor, reads its current speed or position depending on the
    selected mode.

    The positions or speeds are returned in the same order as the motors were
    given in the ``axes`` argument.
    """

    ret = [time()]

    for axis in self._axes:

      # Selecting the register to read depending on the selected mode
      if self._mode == 'speed':
        reg = 10 * axis + input_reg_addr["act_speed"]
      else:
        reg = 10 * axis + input_reg_addr["act_position"]

      # Actually reading the data and converting it
      read = self._variator.read_input_registers(address=reg, count=2)
      converted, *_ = unpack("=f", pack("=HH", *read.registers[::-1]))

      ret.append(converted)

    return ret

  def set_cmd(self, *cmd: float) -> None:
    """Sets either the speed or the position of all motors depending on the
    selected mode.

    If more commands than motors are given, the extra commands are ignored. If
    there are more motors than commands, only part of the motors will be set.
    """

    for axis, val in zip(self._axes, cmd):

      # Setting the speed of each motor
      if self._mode == 'speed':
        reg = 10 * axis + reg_addr['velocity']
        self._variator.write_register(reg, abs(val))

        direction = 10 * axis + reg_addr['direction']
        self._variator.write_register(direction, 0 if val > 0 else 1)

      # Setting the target position of each motor
      else:
        coil = 10 * axis + coil_addr['move_abs']
        reg = 10 * axis + reg_addr['position']
        data = unpack("=HH", pack("=f", val))[::-1]

        self._variator.write_registers(reg, list(data))
        self._variator.write_coil(coil, True)

  def close(self) -> None:
    """Closes the modbus connection to the variator."""

    self._variator.close()



# coding: utf-8

from time import time
from typing import Callable, Union, Dict, Optional
from re import split, IGNORECASE, match

condition_type = Callable[[Dict[str, list]], bool]


class Path:
  """Parent class for all the generator paths.

  Allows them to have access to the:meth:`parse_condition` method.
  """

  def __init__(self,
               _last_time: float,
               _last_cmd: Optional[float] = None) -> None:
    """Simply sets the arguments."""

    self.t0 = _last_time
    self.last_cmd = _last_cmd if _last_cmd is not None else 0

  def get_cmd(self, _: Dict[str, list]) -> float:
    """If not overridden, simply returns the last_cmd attribute."""

    return self.last_cmd

  def parse_condition(
        self,
        condition: Optional[Union[str, condition_type]]) -> condition_type:
    """This method returns a function allowing to check whether the stop
    condition is met or not.

    Its main use is to parse the conditions given as strings, but it can also
    accept :obj:`None` or a callable as arguments.

    If given as a string, the supported condition types are :
    ::

      '<var> > <threshold>'
      '<var> < <threshold>'
      'delay = <your_delay>'

    With ``<var>``, ``<threshold>`` and ``<your_delay>`` to be replaced
    respectively with the label on which the condition applies, the threshold
    for the condition to become true, and the delay before switching to the
    next path.
    """

    if not isinstance(condition, str):
      # First case, the condition is None
      if condition is None:
        return lambda _: False
      # Second case, the condition is already a Callable
      elif isinstance(condition, Callable):
        return condition

    # Third case, the condition is a string containing '<'
    if '<' in condition:
      var, thresh = split(r'\s*<\s*', condition)

      # Return a function that checks if received data is inferior to threshold
      def cond(data: Dict[str, list]) -> bool:
        if var in data:
          return any((val < float(thresh) for val in data[var]))
        return False

      return cond

    # Fourth case, the condition is a string containing '>'
    elif '>' in condition:
      var, thresh = split(r'\s*>\s*', condition)

      # Return a function that checks if received data is superior to threshold
      def cond(data: Dict[str, list]) -> bool:
        if var in data:
          return any((val > float(thresh) for val in data[var]))
        return False

      return cond

    # Fifth case, it is a delay condition
    elif match(r'delay', condition, IGNORECASE) is not None:
      delay = float(split(r'=\s*', condition)[1])
      # Return a function that checks if the delay is expired
      return lambda _: time() - self.t0 > delay

    # Otherwise, it's an invalid syntax
    else:
      raise ValueError("Wrong syntax for the condition, please refer to the "
                       "documentation")



# coding: utf-8

from collections import namedtuple
from struct import unpack
from typing import Union, List, Tuple, Optional, Callable
from _io import FileIO
from multiprocessing.synchronize import RLock
from time import time, sleep

from .ft232h import ft232h
from ..._global import OptionalModule
try:
  from usb import util
  from usb.core import find, USBError

  Ftdi_req_out = util.build_request_type(util.CTRL_OUT,
                                         util.CTRL_TYPE_VENDOR,
                                         util.CTRL_RECIPIENT_DEVICE)

  Ftdi_req_in = util.build_request_type(util.CTRL_IN,
                                        util.CTRL_TYPE_VENDOR,
                                        util.CTRL_RECIPIENT_DEVICE)

except (ModuleNotFoundError, ImportError):
  util = OptionalModule("pyusb")
  find = OptionalModule("pyusb")
  USBError = OptionalModule("pyusb")
  Ftdi_req_out = 0x40
  Ftdi_req_in = 0xC0

ft232h_cmds = {'write_bytes_PVE_MSB': 0x10,
               'write_bytes_NVE_MSB': 0x11,
               'write_bits_NVE_MSB': 0x13,
               'write_bytes_PVE_LSB': 0x18,
               'write_bytes_NVE_LSB': 0x19,
               'read_bytes_PVE_MSB': 0x20,
               'read_bits_PVE_MSB': 0x22,
               'read_bytes_NVE_MSB': 0x24,
               'read_bytes_PVE_LSB': 0x28,
               'read_bytes_NVE_LSB': 0x2C,
               'rw_bytes_PVE_NVE_MSB': 0x31,
               'rw_bytes_NVE_PVE_MSB': 0x34,
               'rw_bytes_PVE_NVE_LSB': 0x39,
               'rw_bytes_NVE_PVE_LSB': 0x3C,
               'set_bits_low': 0x80,
               'get_bits_low': 0x81,
               'set_bits_high': 0x82,
               'get_bits_high': 0x83,
               'loopback_start': 0x84,
               'loopback_end': 0x85,
               'set_tck_divisor': 0x86,
               'send_immediate': 0x87,
               'disable_clk_div5': 0x8A,
               'enable_clk_div5': 0x8B,
               'enable_clk_3phase': 0x8C,
               'disable_clk_3phase': 0x8D,
               'enable_clk_adaptative': 0x96,
               'disable_clk_adaptative': 0x97,
               'drive_zero': 0x9E}

ft232h_sio_req = {'reset': 0x00,
                  'set_event_char': 0x06,
                  'set_error_char': 0x07,
                  'set_latency_timer': 0x09,
                  'set_bitmode': 0x0B,
                  'read_eeprom': 0x90,
                  'write_eeprom': 0x91}

ft232h_sio_args = {'reset': 0,
                   'purge_RX': 1,
                   'purge_TX': 2}

ft232h_eeprom_size = 256
ft232h_tx_empty_bits = 0x60

ft232h_eeprom = {'has_serial_pos': 0x0A,
                 'str_table': 0x0E,
                 'str_position': 0xA0}

ft232h_pins = {'SCL': 0x01,
               'SDAO': 0x02,
               'SDAI': 0x04,
               'SCL_FB': 0x80,
               'SCK': 0x01,
               'DO': 0x02,
               'DI': 0x04,
               'CS': 0x08}

ft232h_i2c_timings = namedtuple('I2CTimings',
                                't_hd_sta t_su_sta t_su_sto t_buf')

ft232h_i2c_speed = {100E3: ft232h_i2c_timings(4.0E-6, 4.7E-6, 4.0E-6, 4.7E-6),
                    400E3: ft232h_i2c_timings(0.6E-6, 0.6E-6, 0.6E-6, 1.3E-6),
                    1E6: ft232h_i2c_timings(0.26E-6, 0.26E-6, 0.26E-6, 0.5E-6)}


class ft232h_server(ft232h):
  """A class for controlling FTDI's USB to Serial FT232H.

  This class is very similar to the :class:`ft232h` except it doesn't
  directly instantiate the USB device nor send commands to it directly.
  Instead, the commands are sent to a USB server managing communication with
  the different FT232H devices.

  Communication in SPI and I2C are implemented, along with GPIO control. The
  name of the methods for SPI and I2C communication are those of :mod:`smbus`
  and :mod:`spidev` libraries, in order to facilitate the use and the
  integration in a multi-backend environment. This class also allows to write a
  USB serial number in the EEPROM, as there's no default serial number on the
  chip.

  Note:
    The FT232H does not support clock stretching and this may cause bugs with
    some I2C devices. Lowering the ``i2c_speed`` may solve the problem.

  Important:
    If using Adafruit's board, its `I2C Mode` switch should of course be set to
    the correct value according to the chosen mode.

  Important:
    **Only for Linux users:** In order to drive the FT232H, the appropriate
    udev rule should be set. This can be done using the `udev_rule_setter`
    utility in ``crappy``'s `util` folder. It is also possible to add it
    manually by running:
    ::

      $ echo "SUBSYSTEM==\\"usb\\", ATTR{idVendor}==\\"0403\\", \
MODE=\\"0666\\\"" | sudo tee ftdi.rules > /dev/null 2>&1

    in a shell opened in ``/etc/udev/rules.d``.

  Important:
    For controlling several FT232H from the same computer, it is first
    necessary to set their USB serial numbers. Otherwise, an error will be
    raised. This can be done using the crappy utility
    ``Set_ft232h_serial_nr.py``.
  """

  def __init__(self,
               mode: str,
               block_number: int,
               current_file: FileIO,
               command_file: FileIO,
               answer_file: FileIO,
               block_lock: RLock,
               current_lock: RLock,
               serial_nr: Optional[str] = None,
               i2c_speed: float = 100E3,
               spi_turbo: bool = False) -> None:
    """Checks the argument validity and initializes the device.

    Args:
      mode: The communication mode, can be :
        ::

          'SPI', 'I2C', 'GPIO_only', 'Write_serial_nr'

        GPIOs can be driven in any mode, but faster speeds are achievable in
        `GPIO_only` mode.
      block_number: The index the block driving this ft232h_server instance has
        been assigned.
      current_file: A file in which the index of the block currently allowed to
        drive the USB server is written.
      command_file: A file in which the current command to be executed by the
        USB server is written.
      answer_file: A file in which the answer to the current command is
        written.
      block_lock: A lock assigned to this block only, for signaling the USB
        server when the command has been written in the command_file.
      current_lock: A lock common to all the blocks that allows the one block
        holding it to communicate with the USB server.
      serial_nr (:obj:`str`, optional): The serial number of the FT232H to
        drive. In `Write_serial_nr` mode, the serial number to be written.
      i2c_speed: In I2C mode, the I2C bus clock frequency in Hz. Available
        values are :
        ::

          100E3, 400E3, 1E6

        or any value between `10kHz` and `100kHz`. Lowering below the default
        value may solve I2C clock stretching issues on some devices.

      spi_turbo: Increases the achievable bus speed, but may not work with some
        devices.

    Note:
      - **CS pin**:
        The CS pin for selecting SPI devices is always `D3`. This pin is
        reserved and cannot be used as a GPIO. If you want to drive the CS line
        manually, it is possible not to drive the CS pin by setting the SPI
        parameter :attr:`no_cs` to :obj:`True` and to drive the CS line from a
        GPIO instead.

      - ``mode``:
        It is not possible to simultaneously control slaves over SPI and I2C,
        due to different hardware requirements for the two protocols. Trying to
        do so will most likely raise an error or lead to inconsistent behavior.

        """

    self._block_number = block_number
    self._current_file = current_file
    self._command_file = command_file
    self._answer_file = answer_file
    self._block_lock = block_lock
    self._current_lock = current_lock

    super().__init__(mode=mode, serial_nr=serial_nr, i2c_speed=i2c_speed,
                     spi_turbo=spi_turbo)

    self._nb_attempt_1 = 80
    self._nb_attempt_2 = 2

  def _handle_command(self, command: list) -> bytes:
    """Parses the command and sends it to the device.

    Args:
      command: The :obj:`list` containing the command type and the arguments if
        any.

    Returns:
      The index of the command.
    """

    # Control transfer out
    if command[0] == 'ctrl_transfer_out':
      self._command_file.write(b'00' + b',' +
                               str(command[1]).encode() + b',' +
                               str(command[2]).encode() + b',' +
                               str(command[3]).encode() + b',' +
                               str(command[4]).encode() + b',' +
                               bytes(command[5]) + b',' +
                               str(command[6]).encode())
      cmd = b'00'
    # Control transfer in
    elif command[0] == 'ctrl_transfer_in':
      self._command_file.write(b'01' + b',' +
                               str(command[1]).encode() + b',' +
                               str(command[2]).encode() + b',' +
                               str(command[3]).encode() + b',' +
                               str(command[4]).encode() + b',' +
                               str(command[5]).encode() + b',' +
                               str(command[6]).encode())
      cmd = b'01'
    # Write operation
    elif command[0] == 'write':
      self._command_file.write(b'02' + b',' +
                               str(command[1]).encode() + b',' +
                               bytes(command[2]) + b',' +
                               str(command[3]).encode())
      cmd = b'02'
    # Read operation
    elif command[0] == 'read':
      self._command_file.write(b'03' + b',' +
                               str(command[1]).encode() + b',' +
                               str(command[2]).encode() + b',' +
                               str(command[3]).encode())
      cmd = b'03'
    # Checks whether the kernel driver is active
    # It doesn't actually interact with the device
    elif command[0] == 'is_kernel_driver_active':
      self._command_file.write(b'04' + b',' +
                               str(command[1]).encode())
      cmd = b'04'
    # Detaches the kernel driver
    # It doesn't actually interact with the device
    elif command[0] == 'detach_kernel_driver':
      self._command_file.write(b'05' + b',' +
                               str(command[1]).encode())
      cmd = b'05'
    # Sets the device configuration
    elif command[0] == 'set_configuration':
      self._command_file.write(b'06')
      cmd = b'06'
    # Custom command getting information from the current configuration
    elif command[0] == 'get_active_configuration':
      self._command_file.write(b'07')
      cmd = b'07'
    # Should the block close the device when leaving ?
    # It doesn't actually interact with the device
    elif command[0] == 'close?':
      self._command_file.write(b'08')
      cmd = b'08'
    # Checks whether the internal resources have been released or not
    # It doesn't actually interact with the device
    elif command[0] == '_ctx.handle':
      self._command_file.write(b'09')
      cmd = b'09'
    # Releases the USB interface
    # It doesn't actually interact with the device
    elif command[0] == 'release_interface':
      self._command_file.write(b'10' + b',' +
                               str(command[1]).encode())
      cmd = b'10'
    # Detaches the kernel driver
    # It doesn't actually interact with the device
    elif command[0] == 'attach_kernel_driver':
      self._command_file.write(b'11' + b',' +
                               str(command[1]).encode())
      cmd = b'11'
    # Releases all the resources used by :mod:`pyusb` for a given device
    # It doesn't actually interact with the device
    elif command[0] == 'dispose_resources':
      self._command_file.write(b'12')
      cmd = b'12'
    # Registers a block as gone
    # It doesn't actually interact with the device
    elif command[0] == 'farewell':
      self._command_file.write(b'13')
      cmd = b'13'
    else:
      raise ValueError("Wrong command type !")

    return cmd

  def _send_server(self, command: list) -> Union[int, bytes, None,
                                                 Tuple[int, ...]]:
    """Sends a command to the server and gets the corresponding answer.

    Args:
      command: A :obj:`list` containing the command type as a first element,
        and then the arguments if any.
    """

    raise_kbi = False  # Flag for postponing the raise of the exception
    retries = 3  # Number of retries for acquiring the lock
    while True:
      try:
        # Acquires the lock assigned to this block only
        if not self._block_lock.acquire(timeout=1):
          retries -= 1
          if not retries:
            raise TimeoutError("Couldn't acquire the lock in a reasonable "
                               "delay !")
          continue
        # Acquires the lock common to all blocks, to start communicating with
        # the server
        if self._current_lock.acquire(timeout=1):
          while True:
            release = True  # Should the common lock be released ?
            try:

              # Writing the block number in the file
              self._current_file.seek(0)
              self._current_file.truncate(0)
              self._current_file.write(str(self._block_number).encode())

              self._command_file.seek(0)
              self._command_file.truncate(0)

              # Writes the command and its arguments to the command file
              cmd = self._handle_command(command)

              # Releases the lock to indicate the server that the command is
              # ready
              self._block_lock.release()

              # Waits for the answer to be written in the answer file
              # It prevents the block from re-acquiring the lock it just
              # released
              try:
                t = time()
                while not self._answer_file.tell():
                  sleep(0.00001)
                  if time() - t > 1:
                    raise TimeoutError
              except (KeyboardInterrupt, TimeoutError):
                raise_kbi = True
                self._block_lock.acquire(timeout=1)
                raise

              # When the lock is re-acquired, the server has written the answer
              # in the answer file
              if self._block_lock.acquire(timeout=1):
                # Reading the answer
                self._answer_file.seek(0)
                answer: List[bytes] = self._answer_file.read().split(b',')
                self._answer_file.seek(0)
                self._answer_file.truncate(0)

                # Making sure we're reading the answer corresponding to our
                # command
                if cmd != answer[0]:
                  raise KeyboardInterrupt

                # The different answers have to be handled in various ways
                if command[0] in ['ctrl_transfer_out',
                                  'write',
                                  'is_kernel_driver_active',
                                  'close?',
                                  '_ctx.handle']:
                  return int(answer[1])
                if command[0] in ['ctrl_transfer_in', 'read']:
                  return answer[1]
                elif command[0] == 'get_active_configuration':
                  return tuple(int(rep) for rep in answer[1:])

                return

              else:
                raise TimeoutError("Couldn't acquire the lock in a reasonable "
                                   "delay !")

            except KeyboardInterrupt:
              # When interrupted, resend the same command to the server without
              # releasing the common lock
              raise_kbi = True
              release = False
              continue
            finally:
              # Releasing the locks if needed
              if release:
                try:
                  self._current_lock.release()
                except AssertionError:
                  pass
                try:
                  self._block_lock.release()
                except AssertionError:
                  pass

        else:
          raise TimeoutError("Couldn't acquire the lock in a reasonable "
                             "delay !")

      except KeyboardInterrupt:
        # Still try to send the command despite the KeyboardInterrupt
        # This way the block can stop properly
        continue
      finally:
        # Raise the KeyboardInterrupt that was postponed if needed
        if raise_kbi:
          raise KeyboardInterrupt

  def _initialize(self) -> None:
    """Initializing the FT232H according to the chosen mode.

    The main differences are for the choice of the clock frequency and
    parameters.
    """

    # FT232H properties
    fifo_sizes = (1024, 1024)
    latency = 2

    # I2C properties
    if self._ft232h_mode == 'I2C':
      timings = ft232h_i2c_speed[self._i2c_speed if self._i2c_speed in
                                 ft232h_i2c_speed else 100E3]
      frequency = self._i2c_speed

      self._ck_hd_sta = self._compute_delay_cycles(timings.t_hd_sta)
      self._ck_su_sto = self._compute_delay_cycles(timings.t_su_sto)
      ck_su_sta = self._compute_delay_cycles(timings.t_su_sta)
      ck_buf = self._compute_delay_cycles(timings.t_buf)
      self._ck_idle = max(ck_su_sta, ck_buf)
      self._ck_delay = ck_buf

      self._i2c_mask = ft232h_pins['SCL'] | ft232h_pins['SDAO'] | \
          ft232h_pins['SDAI']
      self._i2c_dir = ft232h_pins['SCL'] | ft232h_pins['SDAO']

    # SPI properties
    elif self._ft232h_mode == 'SPI':
      frequency = 400E3

      self._bits_per_word = 8
      self._cshigh = False
      self._no_cs = False
      self._loop = False
      self._lsbfirst = False
      self._max_speed_hz = 400E3
      self._mode = 0
      self._threewire = False
      self._spi_param_changed = True

      self._cs_bit = ft232h_pins['CS']
      self._spi_dir = self._cs_bit | ft232h_pins['SCK'] | ft232h_pins['DO']
      self._spi_mask = self._cs_bit | ft232h_pins['SCK'] | \
          ft232h_pins['DO'] | ft232h_pins['DI']

    else:
      frequency = 400E3

    # Configuring the USB device, interface and endpoints
    try:
      if self._send_server(['is_kernel_driver_active', 0]):
        self._send_server(['detach_kernel_driver', 0])
      self._send_server(['set_configuration'])
    except USBError:
      print("You may have to install the udev-rules for this USB device, "
            "this can be done using the udev_rule_setter utility in the util "
            "folder")
      raise

    self._index, self._in_ep, self._out_ep, self._max_packet_size = \
        self._send_server(['get_active_configuration'])

    # Invalidate data in the readbuffer
    self._readoffset = 0
    self._readbuffer = bytearray()
    # Drain input buffer
    self._purge_buffers()
    # Shallow reset

    if self._ctrl_transfer_out(ft232h_sio_req['reset'],
                               ft232h_sio_args['reset']):
      raise IOError('Unable to reset FTDI device')
    # Reset feature mode
    self._set_bitmode(0, ft232h.BitMode.RESET)

    # Set latency timer
    self._set_latency_timer(latency)

    # Set chunk size and invalidate all remaining data
    self._writebuffer_chunksize = fifo_sizes[0]
    self._readoffset = 0
    self._readbuffer = bytearray()
    self._readbuffer_chunksize = min(fifo_sizes[0], fifo_sizes[1],
                                     self._max_packet_size)

    # Reset feature mode
    self._set_bitmode(0, ft232h.BitMode.RESET)
    # Drain buffers
    self._purge_buffers()
    # Disable event and error characters
    if self._ctrl_transfer_out(ft232h_sio_req['set_event_char'], 0):
      raise IOError('Unable to set event char')
    if self._ctrl_transfer_out(ft232h_sio_req['set_error_char'], 0):
      raise IOError('Unable to set error char')

    # Enable MPSSE mode
    if self._ft232h_mode == 'GPIO_only':
      self._set_bitmode(0xFF, ft232h.BitMode.MPSSE)
    else:
      self._set_bitmode(self._direction, ft232h.BitMode.MPSSE)

    # Configure clock
    if self._ft232h_mode == 'I2C':
      # Note that bus frequency may differ from clock frequency, when
      # 3-phase clock is enabled
      self._set_frequency(3 * frequency / 2)
    else:
      self._set_frequency(frequency)

    # Configure pins
    if self._ft232h_mode == 'I2C':
      cmd = bytearray(self._idle)
      cmd.extend((ft232h_cmds['set_bits_high'], 0, 0))
      self._write_data(cmd)
    elif self._ft232h_mode == 'SPI':
      cmd = bytearray((ft232h_cmds['set_bits_low'],
                       self._cs_bit & 0xFF,
                       self._direction & 0xFF))
      cmd.extend((ft232h_cmds['set_bits_high'],
                  (self._cs_bit >> 8) & 0xFF,
                  (self._direction >> 8) & 0xFF))
      self._write_data(cmd)
    else:
      cmd = bytearray((ft232h_cmds['set_bits_low'], 0, 0))
      cmd.extend((ft232h_cmds['set_bits_high'], 0, 0))
      self._write_data(cmd)

    # Disable loopback
    self._write_data(bytearray((ft232h_cmds['loopback_end'],)))
    # Validate MPSSE
    bytes_ = bytes(self._read_data_bytes(2))
    if (len(bytes_) >= 2) and (bytes_[0] == '\xfa'):
      raise IOError("Invalid command @ %d" % bytes_[1])

    # I2C-specific settings
    if self._ft232h_mode == 'I2C':
      self._tx_size, self._rx_size = fifo_sizes

      # Enable 3-phase clock
      self._write_data(bytearray([True and
                                  ft232h_cmds['enable_clk_3phase'] or
                                  ft232h_cmds['disable_clk_3phase']]))

      # Enable drivezero mode
      self._write_data(bytearray([ft232h_cmds['drive_zero'],
                                  self._i2c_mask & 0xFF,
                                  (self._i2c_mask >> 8) & 0xFF]))

    # Disable adaptative clock
    self._write_data(bytearray([False and
                                ft232h_cmds['enable_clk_adaptative'] or
                                ft232h_cmds['disable_clk_adaptative']]))

  def _ctrl_transfer_out(self,
                         reqtype: int,
                         value: int,
                         data: bytes = b'') -> int:
    """Sends a control message to the device.

    Args:
      reqtype (:obj:`int`): bmRequest
      value (:obj:`int`): wValue
      data (:obj:`bytes`): payload

    Returns:
      Number of bytes actually written
    """

    try:
      return self._send_server(['ctrl_transfer_out', Ftdi_req_out, reqtype,
                                value, self._index, bytearray(data),
                                self._usb_write_timeout])
    except USBError as ex:
      raise IOError('UsbError: %s' % str(ex))

  def _read_data_bytes(self,
                       size: int,
                       attempt: int = 2,
                       request_gen: Optional[
                         Callable[[int], Union[bytearray,
                                               bytes]]] = None) -> bytes:
    """Reads data from the FT232H.
    Reads data from the FTDI interface. The data buffer is rebuilt from
    chunk-sized blocks received over the USB bus. The FTDI device always sends
    internal status bytes, which are stripped out as not part of the data
    payload.
    Args:
      size (:obj:`int`): The number of bytes to receive from the device
      attempt (:obj:`int`): Attempt cycle count
      request_gen: A callable that takes the number of bytes read and expects a
        bytes buffer to send back to the remote device. This is only useful to
        perform optimized/continuous transfer from a slave device.
    Returns:
      Payload bytes
    """

    # Packet size sanity check
    if not self._max_packet_size:
      raise ValueError("max_packet_size is bogus")
    packet_size = self._max_packet_size
    length = 1  # initial condition to enter the usb_read loop
    data = bytearray()
    # everything we want is still in the cache?
    if size <= len(self._readbuffer) - self._readoffset:
      data = self._readbuffer[self._readoffset:self._readoffset + size]
      self._readoffset += size
      return data
    # something still in the cache, but not enough to satisfy 'size'?
    if len(self._readbuffer) - self._readoffset != 0:
      data = self._readbuffer[self._readoffset:]
      # end of readbuffer reached
      self._readoffset = len(self._readbuffer)
    # read from USB, filling in the local cache as it is empty
    retry = attempt
    req_size = size
    try:
      while (len(data) < size) and (length > 0):
        while True:

          try:
            tempbuf = self._send_server(['read', self._out_ep,
                                         self._readbuffer_chunksize,
                                         self._usb_read_timeout])
          except USBError:
            raise

          retry -= 1
          length = len(tempbuf)
          # the received buffer contains at least one useful databyte
          # (first 2 bytes in each packet represent the current modem
          # status)
          if length >= 2:
            if tempbuf[1] & ft232h_tx_empty_bits:
              if request_gen:
                req_size -= length - 2
                if req_size > 0:
                  cmd = request_gen(req_size)
                  if cmd:
                    self._write_data(cmd)
          if length > 2:
            retry = attempt
            # skip the status bytes
            chunks = (length + packet_size - 1) // packet_size
            count = packet_size - 2
            self._readbuffer = bytearray()
            self._readoffset = 0
            srcoff = 2
            for _ in range(chunks):
              self._readbuffer += tempbuf[srcoff:srcoff + count]
              srcoff += packet_size
            length = len(self._readbuffer)
            break
          # received buffer only contains the modem status bytes
          # no data received, may be late, try again
          if retry > 0:
            continue
          # no actual data
          self._readbuffer = bytearray()
          self._readoffset = 0
          # no more data to read?
          return data
        if length > 0:
          # data still fits in buf?
          if (len(data) + length) <= size:
            data += self._readbuffer[self._readoffset:
                                     self._readoffset + length]
            self._readoffset += length
            # did we read exactly the right amount of bytes?
            if len(data) == size:
              return data
          else:
            # partial copy, not enough bytes in the local cache to
            # fulfill the request
            part_size = min(size - len(data),
                            len(self._readbuffer) - self._readoffset)
            if part_size < 0:
              raise ValueError("Internal Error")
            data += self._readbuffer[self._readoffset:
                                     self._readoffset + part_size]
            self._readoffset += part_size
            return data
    except USBError:
      print("An error occurred while writing to USB")
      raise
    # never reached
    raise ValueError("Internal error")

  def _set_serial_number(self, serial_number: str) -> None:
    """(Over)Writes the serial number.

    Writes the desired serial number to the EEPROM. It is then accessible to
    USB commands as a string descriptor. Also sets the manufacturer and
    product string descriptors.

    Args:
      serial_number (:obj:`str`): Serial number to be written in the EEPROM
    """

    if not isinstance(serial_number, str):
      serial_number = str(serial_number)
    if any(char in serial_number for char in ':/'):
      raise ValueError("Invalid character : or / in serial number")

    # Reading current eeprom
    word_count = round(ft232h_eeprom_size / 2)
    word_addr = 0
    data = bytearray()
    while word_count:
      try:
        buf = self._send_server(['ctrl_transfer_in', Ftdi_req_in,
                                 ft232h_sio_req['read_eeprom'], 0, word_addr,
                                 2, self._usb_read_timeout])
      except USBError as exc:
        raise IOError('UsbError: %s' % exc) from exc
      if not buf:
        raise IOError('EEPROM read error @ %d' % (word_addr << 1))
      data.extend(buf)
      word_count -= 1
      word_addr += 1
    new_eeprom = data[0:ft232h_eeprom_size]

    # Setting the has_serial flag to True
    new_eeprom[ft232h_eeprom['has_serial_pos']] |= 1 << 3

    # Changing the string descriptors and the descriptors index
    str_descriptors = {'manufacturer': 'FTDI',
                       'product': 'FT232H',
                       'serial': serial_number}
    stream = bytearray()
    str_pos = ft232h_eeprom['str_position']
    tbl_pos = ft232h_eeprom['str_table']
    data_pos = str_pos
    for name in str_descriptors:
      new_str = str_descriptors[name].encode('utf-16le')
      length = len(new_str) + 2
      stream.append(length)
      stream.append(util.DESC_TYPE_STRING)  # string descriptor
      stream.extend(new_str)
      new_eeprom[tbl_pos] = data_pos
      tbl_pos += 1
      new_eeprom[tbl_pos] = length
      tbl_pos += 1
      data_pos += length
    new_eeprom[str_pos:str_pos + len(stream)] = stream

    # Filling the remaining space with zeros
    crc_pos = len(new_eeprom)
    rem = crc_pos - (str_pos + len(stream))
    new_eeprom[str_pos + len(stream):crc_pos] = bytes(rem)

    # Checking the eeprom length
    if len(new_eeprom) != ft232h_eeprom_size:
      raise ValueError("Eeprom_size not matching, serial number may be "
                       "too long, eeprom not written")

    # Calculating the new checksum and modifying the corresponding bytes
    checksum = 0xAAAA
    for idx in range(0, len(new_eeprom[:-2]), 2):
      v = ((new_eeprom[:-2][idx + 1] << 8) + new_eeprom[:-2][idx]) & 0xFFFF
      checksum = v ^ checksum
      checksum = ((checksum << 1) & 0xFFFF) | ((checksum >> 15) & 0xFFFF)

    new_eeprom[-2] = checksum & 0xFF
    new_eeprom[-1] = checksum >> 8

    # Updating the eeprom
    addr = 0
    for word in unpack('<%dH' % (len(new_eeprom) // 2), new_eeprom):
      out = self._send_server(['ctrl_transfer_out', Ftdi_req_out,
                               ft232h_sio_req['write_eeprom'], word, addr >> 1,
                               b'', self._usb_write_timeout])
      if out:
        raise IOError('EEPROM Write Error @ %d' % addr)
      addr += 2

  def _write_data(self, data: Union[bytearray, bytes]) -> int:
    """Writes data to the FT232H.

    Writes the sequence of MPSSE commands and data to the FTDI port. Data
    buffer is split into chunk-sized blocks before being sent over the USB bus.

    Args:
      data: The byte stream to send to the FTDI interface

    Returns:
      Count of written bytes
    """

    offset = 0
    size = len(data)
    try:
      while offset < size:
        write_size = self._writebuffer_chunksize
        if offset + write_size > size:
          write_size = size - offset

        try:
          length = self._send_server(['write', self._in_ep,
                                      data[offset:offset + write_size],
                                      self._usb_write_timeout])
        except USBError:
          raise

        if length <= 0:
          raise USBError("Usb bulk write error")
        offset += length
      return offset
    except USBError:
      print("An error occurred while writing to USB")
      raise

  def close(self) -> None:
    """Closes the FTDI interface/port."""

    if self._send_server(['close?']):

      if self._send_server(['_ctx.handle']):
        try:
          self._set_bitmode(0, ft232h.BitMode.RESET)
          self._send_server(['release_interface', self._index - 1])
        except (IOError, ValueError, USBError):
          pass
        try:
          self._send_server(['attach_kernel_driver', self._index - 1])
        except (NotImplementedError, USBError):
          pass
      self._send_server(['dispose_resources'])

    self._send_server(['farewell'])



# coding: utf-8

"""
Demonstrates how to use modifiers.

Number of modifiers are already defined in ``crappy.modifiers``, but it can
also be a function or any class containing the ``.evaluate()`` method.

No hardware required.
"""

import crappy


# Example of class used as a Modifier
class My_offset_modifier:
  def __init__(self, offset):
    self.offset = offset

  def evaluate(self, data):
    """Method returning the modified values.

    Remember: data is ALWAYS a :obj:`dict`.
    Returning :obj:`None` will drop the data.
    """

    for k in data:
      if k != 't(s)':  # Move everything except the time
        data[k] += self.offset
    return data  # Do not forget to return it!


# Example of function used as a modifier
def mul_by_10(data):
  data['cmd'] *= 10
  return data


if __name__ == "__main__":
  generator = crappy.blocks.Generator(path=[
    {'type': 'constant', 'value': 0, 'condition': 'delay=2'},
    {'type': 'constant', 'value': 1, 'condition': 'delay=2'}
      ] * 20, spam=True)
  graph = crappy.blocks.Grapher(('t(s)', 'cmd'))
  smooth_graph = crappy.blocks.Grapher(('t(s)', 'cmd'))

  crappy.link(generator, graph)
  # We add a moving average to smooth the data
  # and our custom condition that adds and offset of 5
  crappy.link(generator, smooth_graph,
              # The modifiers will be applied in the order of the list
              modifier=[
                # Integrated modifier, will average the values on 100 points
                crappy.modifier.Moving_avg(100),
                # Will add an offset
                My_offset_modifier(5),
                # Will multiply the result by 10
                mul_by_10])

  # This block will simply print "Triggered" followed by the received data
  r = crappy.blocks.Reader('Triggered')

  # Only forward data when the label "cycle" changed its value
  crappy.link(generator, r,
              modifier=crappy.modifier.Trig_on_change('cycle'))

  crappy.start()



# coding: utf-8

import crappy

g1 = crappy.blocks.Generator([{'type': 'protection', 'condition2': 'F1>5',
                               'value1': 10, 'condition1': 'F1<-5',
                               'value2': -10}], cmd_label='v1')

g2 = crappy.blocks.Generator([{'type': 'protection', 'condition2': 'F2>5',
                               'value1': 10, 'condition1': 'F2<-5',
                               'value2': -10}], cmd_label='v2')

g3 = crappy.blocks.Generator([{'type': 'protection', 'condition2': 'F3>5',
                               'value1': 10, 'condition1': 'F3<-5',
                               'value2': -10}], cmd_label='v3')

g4 = crappy.blocks.Generator([{'type': 'protection', 'condition2': 'F4>5',
                               'value1': 10, 'condition1': 'F4<-5',
                               'value2': -10}], cmd_label='v4')

m1 = {'port': '/dev/ttyS4', 'cmd': 'v1'}
m2 = {'port': '/dev/ttyS5', 'cmd': 'v2'}
m3 = {'port': '/dev/ttyS6', 'cmd': 'v3'}
m4 = {'port': '/dev/ttyS7', 'cmd': 'v4'}
common = {'type': 'biaxe'}

m = crappy.blocks.Machine([m1, m2, m3, m4], common)
s = crappy.blocks.IOBlock('Comedi', channels=[1, 2, 3, 4], gain=3749,
                          labels=['t(s)', 'F1', 'F2', 'F3', 'F4'])

g = crappy.blocks.Grapher(('t(s)', 'F1'), ('t(s)', 'F2'),
                          ('t(s)', 'F3'), ('t(s)', 'F4'))

crappy.link(s, g)
crappy.link(s, g1)
crappy.link(s, g2)
crappy.link(s, g3)
crappy.link(s, g4)
crappy.link(g1, m)
crappy.link(g2, m)
crappy.link(g3, m)
crappy.link(g4, m)

crappy.start()



# coding: utf-8

"""
Demonstration of how to create a custom in InOut in Crappy.

This InOut is intended to be used as a template, it doesn't actually read data
from any real device.

No hardware required.
"""

import crappy
import numpy as np
from time import time, sleep

# This class can be used as a starting point to create a new InOut object.
# To add it to crappy, make the imports relative (refer to any other inout),
# move the class to a file in crappy/inout and add the corresponding line
# in crappy/inout/__init__.py


class My_inout(crappy.inout.InOut):
  """A basic example of an InOut object."""

  def __init__(self, value, noisestd=.01):
    # Do not forget to init InOut !
    super().__init__()
    self.value = value
    self.noisestd = noisestd

  def open(self):
    print("Opening device...")
    sleep(1)
    print("Device opened!")

  def close(self):
    print("Closing device...")
    sleep(.5)
    print("Device closed")

  def get_data(self):
    """This is the method that will be called to read the data.

    It must return a list and the first value is always the time (as the number
    of seconds elapsed since 01/01/1970).
    It can return one or more other values, as long as it matches the number of
    labels in IOBlock.
    """

    return [time(), self.value + np.random.normal() * self.noisestd]


if __name__ == '__main__':
  io = crappy.blocks.IOBlock('My_inout', value=1,
                             labels=['t(s)', 'value'], freq=100)

  graph = crappy.blocks.Grapher(('t(s)', 'value'))

  crappy.link(io, graph)
  crappy.start()



# coding: utf-8

from typing import Optional

from .actuator import Actuator
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


class CM_drive(Actuator):
  """Open a new default serial port for communication with a CMdrive actuator.
  """

  def __init__(self, port: str = '/dev/ttyUSB0', baudrate: int = 9600) -> None:
    """Sets the instance attributes.

    Args:
      port (:obj:`str`, optional): Path to connect to the serial port.
      baudrate (:obj:`int`, optional): Set the corresponding baud rate.
    """

    Actuator.__init__(self)
    self.port = port
    self.baudrate = baudrate

  def open(self) -> None:
    self.ser = serial.Serial(self.port, self.baudrate)

  def stop(self) -> None:
    """Stop the motor motion."""

    # close serial connection before to avoid errors
    self.ser.close()
    self.ser.open()
    self.ser.write('SL 0\r')
    # self.ser.readline()
    self.ser.close()

  def reset(self) -> int:
    """Reset the serial communication, before reopening it to set displacement
    to zero."""

    self.ser.close()
    self.ser.open()  # open serial port
    result = input("Reset the system ? This will erase recorded trajectories")
    # send request to the user if he would reset the system
    if result.lower()[0] in ['y', 'o']:
      # send 'DIS' ASCII characters to disable the motor
      self.ser.write('DIS\r')
      # send 'SAVE' ASCII characters to SAVE servostar values
      self.ser.write('SAVE\r')
      # send 'COLDSTART' ASCII characters to reboot servostar
      self.ser.write('COLDSTART\r')
      k = 0
      # print different stages of booting
      while k < 24:
        print(self.ser.readline())
        k += 1
      # self.ser.close() #close serial connection
      return 1
    else:
      # self.ser.close() #close serial connection
      return 0

  def close(self) -> None:
    """Close the designated port."""

    self.stop()
    self.ser.close()

  def clear_errors(self) -> None:
    """Reset errors."""

    self.ser.write("CLRFAULT\r\n")
    self.ser.write("OPMODE 0\r\n EN\r\n")

  def set_speed(self, speed: float) -> None:
    """Pilot in speed mode, requires speed in `mm/min`."""

    self.ser.close()  # close serial connection before to avoid errors
    self.ser.open()  # open serial port
    # velocity = input ('Velocity: \n')#request to the user about velocity
    if abs(speed) < 1000000:
      # send ASCII characters to the servostar to apply velocity task
      self.ser.write('SL ' + str(int(speed)) + '\r')
      self.ser.read(self.ser.inWaiting())
    else:
      print('Maximum speed exceeded')
    self.ser.close()  # close serial connection

  def set_position(self,
                   position: float,
                   _: Optional[float] = None,
                   motion_type: str = 'relative') -> None:
    """Pilot in position mode, needs speed and final position to run
    (in `mm/min` and `mm`)."""

    self.ser.close()  # close serial connection before to avoid errors
    self.ser.open()  # open serial port

    if motion_type == 'relative':
      # send ASCII characters to apply the selected motion task
      self.ser.write('MR %i\r' % position)
    if motion_type == 'absolute':
      # send ASCII characters to apply the selected motion task
      self.ser.write('MA %i\r' % position)
    self.ser.readline()
    self.ser.close()  # close serial connection

  def move_home(self) -> None:
    """Reset the position to zero."""

    self.ser.open()  # open serial port
    # send 'MH' ASCII characters for requesting to the motor to return at pos 0
    self.ser.write('MA 0\r')
    # self.ser.readline()
    self.ser.close()  # close serial connection

  def get_position(self) -> float:
    """Search for the physical position of the motor.

    Returns:
      Physical position of the motor.
    """

    self.ser.close()
    # ser=self.setConnection(self.myPort, self.baudrate)
    # initialise serial port
    self.ser.open()
    # send 'PFB' ASCII characters to request the location of the motor
    self.ser.write('PR P \r')
    pfb = self.ser.readline()  # read serial data from the buffer
    pfb1 = self.ser.readline()  # read serial data from the buffer
    print('%s %i' % (pfb, (int(pfb1))))  # print location
    print('\n')
    self.ser.close()  # close serial connection
    return int(pfb1)



# coding: utf-8

"""
Short code to read a channel on a chosen DAQ board.

Will show the supported boards and ask the user the picked one. Then reads
continuously the first channel.
Note that no additional argument is specified to the DAQ board, so the channel,
the rate and precision are the default values.

Required hardware:
  - Any supported DAQ board
"""

import crappy

if __name__ == "__main__":
  lst = list(crappy.inout.inout_dict.keys())
  for i, c in enumerate(lst):
    print(i, c)
  name = lst[int(input("What board do you want to use ?> "))]

  m = crappy.blocks.IOBlock(name, labels=['t(s)', 'chan0'], verbose=True)

  g = crappy.blocks.Grapher(('t(s)', 'chan0'))

  crappy.link(m, g)

  crappy.start()



# coding: utf-8

from typing import Optional
from .block import Block


class Reader(Block):
  """Reads and prints the data flowing through the input :ref:`Link`."""

  _index = 0

  def __init__(self,
               name: Optional[str] = None,
               freq: float = 50,
               verbose: bool = False) -> None:
    """Sets the arg and initializes the parent class.

    Args:
      name: If set, will be printed to identify the reader.
      freq: The block will try to loop at this frequency.
      verbose: If :obj:`True`, the looping frequency will be printed every 2s.
    """

    Block.__init__(self)
    self.freq = freq
    self.verbose = verbose

    index = self._instance_index()
    self._name = name if name is not None else f'Reader {index}'

  def loop(self) -> None:
    """Simply flushes the link and prints its data."""

    for link in self.inputs:
      data = link.recv_chunk(blocking=False)
      if data is not None:
        for dic in (dict(i) for i in zip(*([(key, value) for value in values]
                                           for key, values in data.items()))):
          print(f'{self._name} got: {dic}')

  @classmethod
  def _instance_index(cls) -> int:
    """Returns the index of the current instance."""

    cls._index += 1
    return cls._index



# coding:utf-8

import crappy
from time import sleep, time

"""
This code provides a not-so-stupid method to have an acceptable values for PID 
gains on an output. This will require the ability to send a command to the 
actuator in open loop and this program asserts that it is a non-integrating and 
stable process.

It will give a step setpoint and analyse the response to try to deduce the
gains. Make sure the step size is representative of the utilisation domain (and 
in range) of the actuator.

Warning:
  Make sure to review the constants and to have an idea of how this program 
  will perform before running it !
"""


actuator_kwargs = {"kv": 1200, "inertia": 3, "rv": .5, "fv": 2e-5}
actuator_class = crappy.actuator.Fake_motor
step = 5
loop_delay = .01


def stable(data):
  if len(data) < 100:
    return False
  return all([1.05 * data[-1] > data[i] > .95 * data[-1]
              for i in range(-len(data) // 3, 0)])


graph = crappy.blocks.Grapher(('t(s)', 'speed'))
link = crappy.links.Link()
graph.inputs.append(link)
# graph.start()
graph.launch(time())
act = actuator_class(**actuator_kwargs)
act.open()
v0 = act.get_speed()

act.set_speed(step)  # replace with set_position if needed

d = []
while True:
  speed = act.get_speed()
  print("speed=", speed)
  link.send({'t(s)': time(), 'speed': speed})
  d.append(speed)
  if stable(d):
    break
  sleep(loop_delay)

print("len(data)=", len(d))
v1 = sum(d[-len(d) // 10:]) / (len(d) // 10)
k = (v1 - v0) / step
print("v0=", v0, "v1=", v1, "k=", k)
a0 = sum([v1 - i for i in d]) * loop_delay
tar = a0 / (v1 - v0)
print("A0=", a0, "Tar=", tar)
a1 = sum([i - v0 for i in d[:int(tar / loop_delay)]]) * loop_delay
print("A1=", a1)
t = 2.7 * a1 / (v1 - v0)
print("T=", t)
L = tar - t
print("L=", L)

kp = (.2 + .45 * t / L) / k
ti = (.4 * L + .8 * t) / (L + .1 * t) * L
td = .5 * L * t / (.3 * L + t)

graph.stop()
print("***** GAINS: ******")
print("P=", kp)
print("I=", ti)
print("D=", td)



# coding: utf-8

import numpy as np
from time import time
from typing import List, Optional
from .inout import InOut
from ..tool import pyspcm as spc


class Spectrum(InOut):
  """This class can read data from a Spectrum high speed ADC interfacing over
  PCIe.

  It can acquire data over multiple channels, and set for each channel a
  different voltage range. It is possible to tune the sample rate, the chunk
  size and the memory allocated to the data buffer.

  This class can only acquire data by streaming, it cannot acquire single data
  points.
  """

  def __init__(self,
               channels: List[int],
               device: str = '/dev/spcm0',
               ranges: Optional[List[int]] = None,
               sample_rate: int = 100000,
               buff_size: int = 2**26,
               notify_size: int = 2**16) -> None:
    """Sets the args and initializes the parent class.

    Args:
      channels: A :obj:`list` of all the channels to read data from, given as
        :obj:`int`. Refer to the documentation to know which combinations of
        channels are allowed.
      device: The address of the device to read data from, as a :obj:`str`.
      ranges: A :obj:`list` indicating for each channel the range of the
        acquisition in mV, as an :obj:`int`. There should be as many values in
        this list as there are channels. If not given, all ranges default to
        `10000` mV.
      sample_rate: The sample rate of the acquisition for all channels, in Hz.
        The default is 100KHz.
      buff_size: The size of the memory allocated as a rolling buffer to copy
        the data from the card, in bytes. The default is 67MB.
      notify_size: The size of each chunk of data to copy from the card, in
        bytes. The default is 65kB.
    """

    super().__init__()

    # Setting the args
    self._channels = channels
    self._device = device.encode()
    self._ranges = ranges if ranges is not None else [10000 for _ in channels]
    self._sample_rate = sample_rate
    self._buff_size = buff_size
    self._notify_size = notify_size
    self._chunk_size = notify_size // (2 * len(channels))

    print(f"[Spectrum] Will send "
          f"{2 * sample_rate * len(channels) / notify_size} chunks of "
          f"{notify_size / 1024} kB per second "
          f"({sample_rate * len(channels) / 512} kB/s)")

    # These attributes will be set later
    self._spectrum = None
    self._dt = None
    self._buff = None
    self._stream_t0 = 0
    self._n_points = 0
    self._stream_started = False

  def open(self) -> None:
    """Opens and configures the Spectrum, and sets the ranges and the sample
    rate as requested."""

    self._spectrum = spc.hOpen(self._device)

    # Configuring the Spectrum
    spc.dw_set_param(self._spectrum, spc.SPC_CHENABLE,
                     sum([2 ** chan for chan in self._channels]))
    spc.dw_set_param(self._spectrum, spc.SPC_CARDMODE, spc.SPC_REC_FIFO_SINGLE)
    spc.dw_set_param(self._spectrum, spc.SPC_TIMEOUT, 5000)
    spc.dw_set_param(self._spectrum, spc.SPC_TRIG_ORMASK,
                     spc.SPC_TMASK_SOFTWARE)
    spc.dw_set_param(self._spectrum, spc.SPC_TRIG_ANDMASK, 0)
    spc.dw_set_param(self._spectrum, spc.SPC_CLOCKMODE, spc.SPC_CM_INTPLL)

    # Setting for each channel its range
    for range_, chan in zip(self._ranges, self._channels):
      spc.dw_set_param(self._spectrum, spc.SPC_AMP0 + 100 * chan, range_)

    # Setting the target sample rate and reading the actual sample rate
    spc.dw_set_param(self._spectrum, spc.SPC_SAMPLERATE, self._sample_rate)
    spc.dw_set_param(self._spectrum, spc.SPC_CLOCKOUT, 0)
    real_samplerate = spc.dw_get_param(self._spectrum, spc.SPC_SAMPLERATE)

    self._dt = 1 / real_samplerate

    # Creating the buffer for data
    self._buff = spc.new_buffer(self._buff_size)
    spc.dw_def_transfer(self._spectrum, spc.SPCM_BUF_DATA,
                        spc.SPCM_DIR_CARDTOPC, self._notify_size, self._buff,
                        0, self._buff_size)

  def start_stream(self) -> None:
    """Starts the streams and saves the corresponding timestamp."""

    spc.dw_set_param(self._spectrum, spc.SPC_M2CMD, spc.M2CMD_CARD_START |
                     spc.M2CMD_CARD_ENABLETRIGGER | spc.M2CMD_DATA_STARTDMA)
    self._stream_t0 = time()
    self._stream_started = True

  def get_stream(self) -> List[np.ndarray]:
    """Waits for data to be available, and returns it along with an array of
    timestamps."""

    # Generating the array of timestamps
    start = self._stream_t0 + self._dt * self._n_points
    t = np.arange(start, start + (self._chunk_size - 1) * self._dt, self._dt)

    # Waiting until data is available
    spc.dw_set_param(self._spectrum, spc.SPC_M2CMD, spc.M2CMD_DATA_WAITDMA)
    pos = spc.dw_get_param(self._spectrum, spc.SPC_DATA_AVAIL_USER_POS)

    # Getting the data
    data = np.frombuffer(self._buff, dtype=np.int16,
                         count=self._notify_size // 2,
                         offset=pos).reshape(
      self._notify_size // (2 * len(self._channels)), len(self._channels))
    ret = data.copy()
    del data

    # Updating the number of received points
    spc.dw_set_param(self._spectrum, spc.SPC_DATA_AVAIL_CARD_LEN,
                     self._notify_size)
    self._n_points += self._chunk_size

    return [t, ret]

  def stop_stream(self) -> None:
    """Stops the stream, if it was started."""

    if self._stream_started:
      spc.dw_set_param(self._spectrum, spc.SPC_M2CMD, spc.M2CMD_CARD_STOP |
                       spc.M2CMD_DATA_STOPDMA)

  def close(self) -> None:
    """Closes the connection to the Spectrum if it was opened."""

    if self._spectrum is not None:
      spc.vClose(self._spectrum)



# coding: utf-8

import numpy as np
from typing import Callable, Optional, Union, List, Dict, Any
from pathlib import Path

from ..tool import DISCorrel as Dis
from ..tool import DISConfig, Box
from .camera import Camera
from .displayer import Displayer


class DISCorrel(Camera):
  """This block performs Dense Inverse Search on an image using OpenCV's
  DISOpticalFlow.

  First, a region of interest has to be selected in a GUI. At each new frame,
  the displacement field between the frame and the reference image is projected
  on the desired base of fields, and the returned results are then the
  average values of each field in the selected region of interest.

  This block is mainly intended for calculating the average displacement and/or
  the average strain in the region of interest, but other fields can also be
  computed.
  """

  def __init__(self,
               camera: str,
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               config: bool = True,
               display_images: bool = False,
               displayer_backend: Optional[str] = None,
               displayer_framerate: float = 5,
               verbose: bool = False,
               freq: float = 200,
               save_images: bool = False,
               img_name: str = "{self._n_loops:6d}_{t-self.t0:.6f}.tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               fields: List[str] = None,
               labels: List[str] = None,
               alpha: float = 3,
               delta: float = 1,
               gamma: float = 0,
               finest_scale: int = 1,
               iterations: int = 1,
               gradient_iterations: int = 10,
               init: bool = True,
               patch_size: int = 8,
               patch_stride: int = 3,
               residual: bool = False,
               **kwargs) -> None:
    """Sets the args and initializes the camera.

    Args:
      camera: The name of the camera to control. See :ref:`Cameras` for an
        exhaustive list of available cameras.
      transform: A function taking an image as an argument and returning a
        transformed image. The original image is discarded and only the
        transformed one is kept for processing, display and saving.
      config: If :obj:`True`, a config window is shown before the test starts
        for interactively tuning the camera settings. It also allows selecting
        the spots to track.
      display_images: If :obj:`True`, a window displays the acquired images
        in low resolution during the test. This display is mainly intended for
        debugging and visual follow-up, but not for displaying high-quality
        images.
      displayer_backend: If ``display_images`` is :obj:`True`, the backend to
        use for the display window. Should be one of :
        ::

          'cv2', 'mpl'

        If not given, OpenCV will be used if available.
      displayer_framerate: If ``display_images`` is :obj:`True`, sets the
        maximum framerate for updating the display window. This setting allows
        limiting the resources used by the displayer. Note that the actual
        achieved framerate might differ, this is just the maximum limit.
      verbose: If :obj:`True`, the achieved framerate will be displayed in the
        console during the test.
      freq: If given, the block will try to loop at this frequency. If it is
        lower than the framerate of the camera, frames will be dropped. This
        argument can be used for limiting the achieved framerate when the
        camera doesn't support framerate control.
      save_images: If :obj:`True`, the acquired images are saved on the
        computer during the test. Note that saving images uses CPU, so the
        achieved performance might drop when this feature is in use.
      img_name: If ``save_images`` is :obj:`True`, the template for naming the
        recorded images. It is evaluated as an `f-string`, and must contain the
        file extension at the end. For building the `f-string`, the
        ``self._n_loops`` attribute holds the loop number, and ``t-self.t0``
        holds the current timestamp.
      save_folder: If ``save_images`` is :obj:`True`, the directory to save
        images to. If it doesn't exist, it will be created. If not given, the
        images are saved in a folder named `Crappy_images` and created next to
        the file being run.
      save_period: If ``save_images`` is :obj:`True`, only one every this
        number of images will be saved.
      save_backend: The backend to use for saving the images. Should be one
        of :
        ::

          'sitk', 'pil', 'cv2'

        If not specified, SimpleITK will be used if available, then OpenCV as a
        second choice, and finally Pillow if none of the others was available.
      image_generator: A function taking two floats as arguments, and returning
        an image. It is only used for demonstration without camera in the
        examples, and isn't meant to be used in an actual test.
      fields: A :obj:`list` of :obj:`str` representing the base of fields on
        which the image will be projected during correlation. The possible
        fields are :
        ::

          'x', 'y', 'r', 'exx', 'eyy', 'exy', 'eyx', 'exy2', 'z'

        If not given, the default fields are :
        ::

          ["x", "y", "exx", "eyy"]

      labels: A :obj:`list` containing the labels to send to downstream blocks.
        The first label should be time, the second the metadata, and there
        should then be one label per field. If not given, the default labels
        are :
        ::

          ['t(s)', 'meta', 'x(pix)', 'y(pix)', 'Exx(%)', 'Eyy(%)']

      alpha: Weight of the smoothness term in DisFlow.
      delta: Weight of the color constancy term in DisFlow.
      gamma: Weight of the gradient constancy term in DisFlow.
      finest_scale: Finest level of the Gaussian pyramid on which the flow
        is computed in DisFlow (`0` means full scale).
      iterations: Maximum number of gradient descent iterations in the
        patch inverse search stage in DisFlow.
      gradient_iterations: Maximum number of gradient descent iterations
        in the patch inverse search stage in DisFlow.
      init: If :obj:`True`, the new optical flow is at each loop initialized
        using the previous optical flow.
      patch_size: Size of an image patch for matching in DisFlow
        (in pixels).
      patch_stride: Stride between neighbor patches in DisFlow. Must be
        less than patch size.
      residual: If :obj:`True`, the residuals are computed and sent under the
        label ``'res'``. This label shouldn't be included if custom labels are
        given.
      **kwargs: Any additional argument to pass to the camera.
    """

    super().__init__(camera=camera,
                     transform=transform,
                     config=False,
                     display_images=False,
                     displayer_backend=displayer_backend,
                     displayer_framerate=displayer_framerate,
                     software_trig_label=None,
                     verbose=verbose,
                     freq=freq,
                     save_images=save_images,
                     img_name=img_name,
                     save_folder=save_folder,
                     save_period=save_period,
                     save_backend=save_backend,
                     image_generator=image_generator,
                     **kwargs)

    self._dis = Dis(fields=fields,
                    alpha=alpha,
                    delta=delta,
                    gamma=gamma,
                    finest_scale=finest_scale,
                    init=init,
                    iterations=iterations,
                    gradient_iterations=gradient_iterations,
                    patch_size=patch_size,
                    patch_stride=patch_stride)

    # Managing the fields and labels lists
    fields = ["x", "y", "exx", "eyy"] if fields is None else fields
    self.labels = ['t(s)', 'meta', 'x(pix)', 'y(pix)',
                   'Exx(%)', 'Eyy(%)'] if labels is None else labels
    if residual:
      self.labels.append('res')

    # Making sure a coherent number of labels and fields was given
    if 2 + len(fields) + int(residual) != len(self.labels):
      raise ValueError("The number of fields is inconsistent with the number "
                       "of labels !\nMake sure that the time label was given")

    self._residual = residual
    self._config_dis = config
    self._bbox = Box()

    # Instantiating the displayer
    if display_images:
      self._displayer_dis = Displayer(f"Displayer {camera} "
                                      f"{Camera.cam_count[self._camera_name]}",
                                      displayer_framerate,
                                      displayer_backend)
    else:
      self._displayer_dis = None

  def prepare(self) -> None:
    """Opening the camera, displaying the config window and setting the region
    of interest."""

    super().prepare()

    if self._config_dis:
      conf = DISConfig(self._camera)
      conf.main()
      self._bbox = conf.box

    if self._bbox.no_points():
      raise AttributeError("The region of interest wasn't properly selected in"
                           " the config window !")

    self._dis.set_box(self._bbox)

    if self._displayer_dis is not None:
      self._displayer_dis.prepare()

  def begin(self) -> None:
    """Capturing a first image and setting it as a reference for the
    correlation."""

    _, img = self._camera.get_image()
    self._dis.set_img0(img)

  def finish(self) -> None:
    """Closing the displayer and the camera."""

    if self._displayer_dis is not None:
      self._displayer_dis.finish()

    super().finish()

  def _additional_loop(self, meta: Dict[str, Any], img: np.ndarray) -> None:
    """Getting the updated values from the correlation and updating the
    displayer."""

    data = self._dis.get_data(img, self._residual)
    self.send([meta['t(s)'], meta] + data)

    if self._displayer_dis is not None:
      self._draw_box(img, self._bbox)
      self._displayer_dis.update(img)



# coding: utf-8

"""
Detailed example showing a simple tensile test.

This example uses a Labjack T7 board to send the position command to a tensile
machine. This was tested on an Instron 5882, but any tensile machine taking an
external command can be used.

Required hardware:
  - Tensile machine able to take an external command
  - DAQ board (here a Labjack T7)
"""

import crappy

if __name__ == "__main__":
  SPEED = 5  # mm/min
  FORCE_GAIN = 100  # N/V
  POS_GAIN = 5  # mm/V

  # The path we will use in the generator
  path = {
    "type": "ramp",
    "speed": SPEED / 60,  # Convert to mm/s (in ramps, speed is always
    # in unit/s)
    "condition": None}  # No exit condition: we will stop the test manually

  # Let's create the first block: the generator
  gen = crappy.blocks.Generator([path], cmd_label='cmd')

  # Now the DAQ board we will interact with. Let's define the channels first
  # First input: the force sensor
  force = {'name': 'AIN0', 'gain': FORCE_GAIN}
  # Second input: the position sensor
  # make_zero asks the board to read the value at the beginning and subtract
  # if from the measured value to make it relative to the beginning of the test
  pos = {'name': 'AIN1', 'gain': POS_GAIN, 'make_zero': True}
  # Only output: the command sent to the machine
  # Because TDACx is necessarily an output and AINx an input, the block will
  # guess the direction of each channel. If ambiguous it must be specified with
  # direction=...  (False for input, True for output)
  cmd = {'name': 'TDAC0', 'gain': POS_GAIN}
  # Now let us create the block: the first arg is the name of the InOut object
  # to use. The block will instantiate it itself ! No need to create it here
  # Then we give the channels, the InOut will automatically detect that it has
  # two inputs and one output
  # The output labels always start with the time in second so we need 3 outputs
  daq = crappy.blocks.IOBlock('Labjack_t7', channels=[force, pos, cmd],
                              labels=['t(s)', 'F(N)', 'Position(mm)'],
                              cmd_labels=['cmd'])
  # Now we link them: the output of the board will be read from the label 'cmd'
  # Note that links are one-way only, so the order is important
  crappy.link(gen, daq)

  # Now we would like to save the measurements from the machine
  rec = crappy.blocks.Recorder('results.csv')
  crappy.link(daq, rec)

  # And why not display the force in real time during the test
  graph = crappy.blocks.Grapher(('t(s)', 'F(N)'))
  crappy.link(daq, graph)

  # Now that all the blocks are created and linked, the test can be started
  # Do not forget to enable the external command on the machine beforehand
  crappy.start()
  # To interrupt the test, simply press ctrl+C in the terminal
  # It will stop all the blocks and end the program



# coding: utf-8

from time import sleep
from typing import Optional

from .._global import DefinitionError


class MetaActuator(type):
  """Metaclass ensuring that two Actuators don't have the same name, and that
  all Actuators define the required methods. Also keeps track of all the
  Actuator classes, including the custom user-defined ones."""

  classes = {}
  needed_methods = ['open', 'stop', 'close']

  def __new__(mcs, name: str, bases: tuple, dct: dict) -> type:
    return super().__new__(mcs, name, bases, dct)

  def __init__(cls, name: str, bases: tuple, dct: dict) -> None:
    super().__init__(name, bases, dct)

    # Checking that an InOut with the same name doesn't already exist
    if name in cls.classes:
      raise DefinitionError(f"The {name} class is already defined !")

    # Gathering all the defined methods
    defined_methods = list(dct.keys())
    defined_methods += [base.__dict__.keys() for base in bases]

    # Checking for missing methods
    missing_methods = [meth for meth in cls.needed_methods
                       if meth not in defined_methods]

    # Raising if there are unexpected missing methods
    if missing_methods and name != "Actuator":
      raise DefinitionError(
        f'Class {name} is missing the required method(s): '
        f'{", ".join(missing_methods)}')

    # Otherwise, saving the class
    if name != 'Actuator':
      cls.classes[name] = cls


class Actuator(metaclass=MetaActuator):
  """The base class for all actuator classes, allowing to keep track of them
  and defining methods shared by all of them."""

  def set_speed(self, speed: float) -> None:
    """This method should drive the actuator so that it reaches the desired
    speed."""

    print(f"WARNING ! Trying to drive the Actuator {type(self).__name__} in "
          f"speed but it does not define a set_speed method !\nNo command "
          f"sent to the actuator.")
    sleep(1)
    return

  def set_position(self,
                   position: float,
                   speed: Optional[float] = None) -> None:
    """This method should drive the actuator so that it reaches the desired
    position. A speed value can optionally be provided for specifying the speed
    at which the actuator should move for getting to the desired position."""

    print(f"WARNING ! Trying to drive the Actuator {type(self).__name__} in "
          f"position but it does not define a set_position method !\n"
          f"No command sent to the actuator.")
    sleep(1)
    return

  def get_speed(self) -> Optional[float]:
    """This method should return the current speed of the actuator. It is also
    fine for this method to return :obj:`None`."""

    print(f"WARNING ! Trying to get the speed of the Actuator "
          f"{type(self).__name__}, but it does not define a get_speed "
          f"method !\nDefine such a method, don't set the speed_label, or "
          f"remove the output links of this block.")
    sleep(1)
    return

  def get_position(self) -> Optional[float]:
    """This method should return the current position of the actuator. It is
    also fine for this method to return :obj:`None`."""

    print(f"WARNING ! Trying to get the position of the Actuator "
          f"{type(self).__name__}, but it does not define a get_position "
          f"method !\nDefine such a method, don't set the pos_label, or remove"
          f" the output links of this block.")
    sleep(1)
    return



# coding: utf-8

from datetime import timedelta
from time import time
from typing import Tuple, List, Dict, Any, Optional

from .block import Block
from .._global import OptionalModule

try:
  import matplotlib.pyplot as plt
  from matplotlib import cm
except (ModuleNotFoundError, ImportError):
  plt = OptionalModule("matplotlib")
  cm = OptionalModule("matplotlib")


class Text:
  """Displays a simple text line on the drawing."""

  def __init__(self,
               _,
               coord: Tuple[int, int],
               text: str,
               label: str,
               **__: str) -> None:
    """Simply sets the args.

    Args:
      _: The parent drawing block.
      coord: The coordinates of the text on the drawing.
      text: The text to display.
      label: The label carrying the information for updating the text.
      **__: Other unused arguments.
    """

    x, y = coord
    self._text = text
    self._label = label

    self._txt = plt.text(x, y, text)

  def update(self, data: Dict[str, float]) -> None:
    """Updates the text according to the received values."""

    self._txt.set_text(self._text % data[self._label])


class Dot_text:
  """Like :class:`Text`, but with a colored dot to visualize a numerical value.
  """

  def __init__(self,
               drawing,
               coord: Tuple[int, int],
               text: str,
               label: str,
               **__: str) -> None:
    """Simply sets the args.

    Args:
      drawing: The parent drawing block.
      coord: The coordinates of the text and the color dot on the drawing.
      text: The text to display.
      label: The label carrying the information for updating the text and the
        color of the dot.
      **__: Other unused arguments.

        Important:
          The value received in label must be a numeric value. It will be
          normalized on the ``crange`` of the block and the dot will change
          color from blue to red depending on this value.
    """

    x, y = coord
    self._text = text
    self._label = label

    self._txt = plt.text(x + 40, y + 20, text, size=16)
    self._dot = plt.Circle(coord, 20)

    drawing.ax.add_artist(self._dot)
    low, high = drawing.color_range

    self._amp = high - low
    self._low = low

  def update(self, data: Dict[str, float]) -> None:
    """Updates the text and the color dot according to the received values."""

    self._txt.set_text(self._text % data[self._label])
    self._dot.set_color(cm.coolwarm((data[self._label] -
                                     self._low) / self._amp))


class Time:
  """Displays a time counter on the drawing, starting at the beginning of the
  test."""

  def __init__(self, drawing, coord: Tuple[int, int], **__) -> None:
    """Simply sets the args.

    Args:
      drawing: The parent drawing block.
      coord: The coordinates of the time counter on the drawing.
      **__: Other unused arguments.
    """

    self._block = drawing
    x, y = coord

    self._txt = plt.text(x, y, "00:00", size=38)

  def update(self, _: Dict[str, float]) -> None:
    """Updates the time counter, independently of the received values."""

    self._txt.set_text(str(timedelta(seconds=int(time() - self._block.t0))))


class Drawing(Block):
  """This block allows displaying a real-time visual representation of data.

  It displays the data on top of a background image and updates it according to
  the values received through the incoming links.

  It is possible to display simple text, a time counter, ot text associated
  with a color dot evolving depending on a predefined color bar and the
  received values.
  """

  def __init__(self,
               image: str,
               draw: Optional[List[Dict[str, Any]]] = None,
               color_range: Tuple[float, float] = (20, 300),
               title: str = "Drawing",
               window_size: Tuple[int, int] = (7, 5),
               backend: str = "TkAgg",
               freq: float = 2,
               verbose: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      image: Path to the image that will be the background of the canvas, as a
        :obj:`str`.
      draw: A :obj:`list` of :obj:`dict` defining what to draw. See below for
        more details.
      color_range: A :obj:`tuple` containing the lowest and highest values for
        the color bar.
      title: The title of the window containing the drawing.
      window_size: The x and y dimension of the window, following
        :mod:`matplotlib` nomenclature.
      backend: The :mod:`matplotlib` backend to use.
      freq: The block will try to loop at this frequency.
      verbose: If :obj:`True`, prints the looping frequency of the block.

    Note:
      - Information about the ``draw`` keys:

        - ``type``: Mandatory, the type of drawing to display. It can be either
          `'text'`, `'dot_text'` or `'time'`.

        - ``coord``: Mandatory, a :obj:`tuple` containing the `x` and `y`
          coordinates where the element should be displayed on the drawing.

        - ``text``: Mandatory for :class:`Text` and :class:`Dot_text` only, the
          text to display on the drawing. It must follow the %-formatting, and
          contain exactly one %-field. This field will be updated using the
          value carried by ``label``.

        - ``label``: Mandatory for :class:`Text` and :class:`Dot_text` only,
          the label of the data to display. It will try to retrieve this data
          in the incoming links. The ``text`` will then be updated with this
          data.
    """

    super().__init__()
    self.freq = freq
    self.verbose = verbose

    self._image = image
    self._draw = [] if draw is None else draw
    self.color_range = color_range
    self._title = title
    self._window_size = window_size
    self._backend = backend

  def prepare(self) -> None:
    """Initializes the different elements of the drawing."""

    # Initializing the window and the background image
    plt.switch_backend(self._backend)
    self._fig, self.ax = plt.subplots(figsize=self._window_size)
    image = self.ax.imshow(plt.imread(self._image), cmap=cm.coolwarm)
    image.set_clim(-0.5, 1)

    # Initializing the color bar
    cbar = self._fig.colorbar(image, ticks=[-0.5, 1], fraction=0.061,
                              orientation='horizontal', pad=0.04)
    cbar.set_label('Temperatures(C)')
    cbar.ax.set_xticklabels(self.color_range)

    # Setting the title and the axes
    self.ax.set_title(self._title)
    self.ax.set_axis_off()

    # Adding the elements to the drawing
    self._drawing_elements = []
    for dic in self._draw:
      if dic['type'] == 'text':
        self._drawing_elements.append(Text(self, **dic))
      elif dic['type'] == 'dot_text':
        self._drawing_elements.append(Dot_text(self, **dic))
      elif dic['type'] == 'time':
        self._drawing_elements.append(Time(self, **dic))

  def loop(self) -> None:
    """Receives the latest data from upstream blocks and updates the drawing
    accordingly."""

    data = self.get_last()
    for elt in self._drawing_elements:
      elt.update(data)
    self._fig.canvas.draw()
    plt.pause(0.001)

  def finish(self) -> None:
    """Simply closes the window containing the drawing."""

    plt.close()
